"""Parse backfill entries."""

import enum
import os
from datetime import date
from pathlib import Path
from typing import List, Optional

import attr
import yaml

from bigquery_etl.query_scheduling.utils import is_email_or_github_identity

BACKFILL_FILE = "backfill.yaml"
DEFAULT_WATCHER = "nobody@mozilla.com"
DEFAULT_REASON = "Please provide a reason for the backfill and links to any related bugzilla or jira tickets"
DEFAULT_BILLING_PROJECT = "moz-fx-data-backfill-slots"


class UniqueKeyLoader(yaml.SafeLoader):
    """YAML loader to check duplicate keys."""

    def construct_mapping(self, node, deep=False):
        """Create mapping while checking for duplicate keys."""
        mapping = set()
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if key in mapping:
                raise ValueError(
                    f"Backfill entry already exists with entry date: {key}."
                )
            mapping.add(key)
        return super().construct_mapping(node, deep)


class Literal(str):
    """Represents a YAML literal."""

    pass


def literal_presenter(dumper, data):
    """Literal representer for YAML output."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.add_representer(Literal, literal_presenter)


class BackfillStatus(enum.Enum):
    """Represents backfill status types."""

    INITIATE = "Initiate"
    COMPLETE = "Complete"


@attr.s(auto_attribs=True)
class Backfill:
    """
    Representation of a backfill entry configuration.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    entry_date: date = attr.ib()
    start_date: date = attr.ib()
    end_date: date = attr.ib()
    excluded_dates: List[date] = attr.ib()
    reason: str = attr.ib()
    watchers: List[str] = attr.ib()
    status: BackfillStatus = attr.ib()
    custom_query_path: Optional[str] = attr.ib(None)
    shredder_mitigation: Optional[bool] = attr.ib(False)
    override_retention_limit: Optional[bool] = attr.ib(False)
    billing_project: Optional[str] = attr.ib(None)

    def __str__(self):
        """Return print friendly string of backfill object."""
        backfill_str = f"""
            entry_date = {self.entry_date}
            start_date = {self.start_date}
            end_date = {self.end_date}
            """.rstrip()

        if self.excluded_dates:
            backfill_str += f"""
            excluded_dates = {[str(e) for e in self.excluded_dates]}
            """.rstrip()

        backfill_str += f"""
            reason = {self.reason}
            watcher(s) = {self.watchers}
            status = {self.status.value}
            custom_query_path = {self.custom_query_path}
            shredder_mitigation = {self.shredder_mitigation}
            override_retention_limit = {self.override_retention_limit}
            """

        return backfill_str.replace("'", "")

    @entry_date.validator
    def validate_entry_date(self, attribute, value):
        """Check that provided entry date is not in the future."""
        if date.today() < value:
            raise ValueError(f"Backfill entry {value} can't be in the future.")

    @start_date.validator
    def validate_start_date(self, attribute, value):
        """Check that provided start date is before end date and entry date."""
        if self.end_date < value or self.entry_date < value:
            raise ValueError(f"Invalid start date: {value}.")

    @end_date.validator
    def validate_end_date(self, attribute, value):
        """Check that provided end date is after start date and before entry date."""
        if value < self.start_date or value > self.entry_date:
            raise ValueError(f"Invalid end date: {value}.")

    @excluded_dates.validator
    def validate_excluded_dates(self, attribute, value):
        """Check that provided excluded dates are between start and end dates, are sorted and contain no duplicates."""
        if not all(map(lambda e: self.start_date < e < self.end_date, value)):
            raise ValueError(f"Invalid excluded dates: {value}.")

        if not value == sorted(value):
            raise ValueError(
                f"Existing backfill entry with excluded dates not sorted: {value}."
            )
        if not len(value) == len(set(value)):
            raise ValueError(
                f"Existing backfill entry with duplicate excluded dates: {value}."
            )

    @watchers.validator
    def validate_watchers(self, attribute, value):
        """Check that provided watchers are valid emails or Github identity with no duplicates."""
        if not value or not all(
            map(lambda e: e and is_email_or_github_identity(e), value)
        ):
            raise ValueError(f"Invalid email or Github identity for watchers: {value}.")

        if len(value) != len(set(value)):
            raise ValueError(f"Duplicate watcher in ({value}).")

    @reason.validator
    def validate_reason(self, attribute, value):
        """Check that provided status is not empty."""
        if not value:
            raise ValueError("Reason in backfill entry should not be empty.")

    @status.validator
    def validate_status(self, attribute, value):
        """Check that provided status is valid."""
        if not hasattr(BackfillStatus, value.name):
            raise ValueError(f"Invalid status: {value.name}.")

    @billing_project.validator
    def validate_billing_project(self, attribute, value):
        """Check that billing project is valid."""
        if value and not value.startswith("moz-fx-data-backfill-"):
            raise ValueError(
                f"Invalid billing project: {value}.  Please use one of the projects assigned to backfills."
            )

    @staticmethod
    def is_backfill_file(file_path: Path) -> bool:
        """Check if the provided file is a backfill file."""
        return os.path.basename(file_path) == BACKFILL_FILE

    @classmethod
    def entries_from_file(
        cls, file: Path, status: Optional[str] = None
    ) -> List["Backfill"]:
        """
        Parse all backfill entries from the provided yaml file.

        Return a list with all backfill entries.

        @param status:  optional status param for filtering backfill entries with specific status.
        If status is not provided, all backfill entries will be returned.
        """
        if not cls.is_backfill_file(file):
            raise ValueError(f"Invalid file: {file}.")

        backfill_entries: List[Backfill] = []

        with open(file, "r") as yaml_stream:
            try:
                backfills = yaml.load(yaml_stream, Loader=UniqueKeyLoader) or {}

                for entry_date, entry in backfills.items():
                    if status is not None and entry["status"] != status:
                        continue

                    backfill = cls(
                        entry_date=entry_date,
                        start_date=entry["start_date"],
                        end_date=entry["end_date"],
                        excluded_dates=entry.get("excluded_dates", []),
                        reason=entry["reason"],
                        watchers=entry["watchers"],
                        status=BackfillStatus[entry["status"].upper()],
                        custom_query_path=entry.get("custom_query_path", None),
                        shredder_mitigation=entry.get("shredder_mitigation", False),
                        override_retention_limit=entry.get(
                            "override_retention_limit", False
                        ),
                        billing_project=entry.get("billing_project", None),
                    )

                    backfill_entries.append(backfill)

            except yaml.YAMLError as e:
                raise ValueError(f"Unable to parse Backfill file {file}") from e
            except ValueError as e:
                raise ValueError(f"Unable to parse Backfill file {file}: {e}") from e

            return backfill_entries

    def to_yaml(self) -> str:
        """Create dictionary version of yaml for writing to file."""
        yaml_dict = {
            self.entry_date: {
                "start_date": self.start_date,
                "end_date": self.end_date,
                "excluded_dates": sorted(self.excluded_dates),
                "reason": self.reason,
                "watchers": self.watchers,
                "status": self.status.value,
                "custom_query_path": self.custom_query_path,
                "shredder_mitigation": self.shredder_mitigation,
                "override_retention_limit": self.override_retention_limit,
                "billing_project": self.billing_project,
            }
        }

        if yaml_dict[self.entry_date]["excluded_dates"] == []:
            del yaml_dict[self.entry_date]["excluded_dates"]

        if yaml_dict[self.entry_date]["custom_query_path"] is None:
            del yaml_dict[self.entry_date]["custom_query_path"]

        if yaml_dict[self.entry_date]["shredder_mitigation"] is None:
            del yaml_dict[self.entry_date]["shredder_mitigation"]

        if yaml_dict[self.entry_date]["override_retention_limit"] is None:
            del yaml_dict[self.entry_date]["override_retention_limit"]

        if yaml_dict[self.entry_date]["billing_project"] is None:
            del yaml_dict[self.entry_date]["billing_project"]

        return yaml.dump(
            yaml_dict,
            sort_keys=False,
        )



from pathlib import Path
from textwrap import dedent

import pytest
from click.testing import CliRunner

from bigquery_etl.cli.check import _build_jinja_parameters, _parse_check_output, _render


class TestCheck:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_parse_check_output(self):
        expected = "ETL Data Check Failed: a check failed"
        assert _parse_check_output(expected) == expected

        test2 = "remove prepended text ETL Data Check Failed: a check failed"
        assert _parse_check_output(test2) == expected

        test3 = "no match for text Data Check Failed: a check failed"
        assert _parse_check_output(test3) == test3

    def test_build_jinja_parameters(self):
        test = [
            "--parameter=submission_date::2023-06-01",
            "--parameter=id::asdf",
            "--use_legacy_sql=false",
            "--project_id=moz-fx-data-marketing-prod",
            "--debug",
        ]
        expected = {
            "submission_date": "2023-06-01",
            "id": "asdf",
            "use_legacy_sql": "false",
            "project_id": "moz-fx-data-marketing-prod",
        }
        assert _build_jinja_parameters(test) == expected

    def test_check_render(self):
        checks_file = Path(
            "tests/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_v6/checks.sql"
        )

        actual = _render(
            checks_file=checks_file,
            dataset_id="telemetry_derived",
            table="clients_daily_v6",
            project_id="moz-fx-data-shared-prod",
            query_arguments=[
                "--parameter=submission_date:DATE:2023-07-01",
            ],
        )

        expected = dedent(
            """
        #fail
        ASSERT(
          (
            SELECT
              COUNT(*)
            FROM
              `moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6`
            WHERE
              submission_date = "2023-07-01"
          ) > 0
        )
        AS
          'ETL Data Check Failed: Table moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6 contains 0 rows for date: 2023-07-01.'"""
        )

        assert actual == expected



import os
from pathlib import Path

import pytest
import yaml
from click.testing import CliRunner

from bigquery_etl.cli.dag import create, generate, info, remove

TEST_DIR = Path(__file__).parent.parent


class TestDag:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_dag_info_without_tasks(self, runner):
        result = runner.invoke(
            info, ["--dags_config=" + str(TEST_DIR / "data" / "dags.yaml")]
        )
        assert result.exit_code == 0
        assert "bqetl_core" in result.output
        assert "bqetl_events" in result.output

    def test_dag_info_with_tasks(self, runner):
        result = runner.invoke(
            info,
            [
                "--with_tasks",
            ],
        )
        assert result.exit_code == 0

    def test_single_dag_info(self, runner):
        result = runner.invoke(
            info,
            ["bqetl_core", "--dags_config=" + str(TEST_DIR / "data" / "dags.yaml")],
        )
        assert result.exit_code == 0
        assert "bqetl_core" in result.output
        assert "bqetl_events" not in result.output

    def test_dag_create_invalid_name(self, runner):
        with runner.isolated_filesystem():
            dags_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                    "tags": ["impact/tier_3"],
                }
            }

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dags_conf))

            result = runner.invoke(
                create,
                [
                    "invalid_dag_name",
                    "--schedule_interval=daily",
                    "--owner=test@example.org",
                    "--description=test",
                    "--start_date=2020-01-01",
                    "--tag=impact/tier_3",
                ],
            )
            assert result.exit_code == 1

    def test_dag_create(self, runner):
        with runner.isolated_filesystem():
            dags_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                    "tags": ["impact/tier_3"],
                }
            }

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dags_conf))

            result = runner.invoke(
                create,
                [
                    "bqetl_new_dag",
                    "--schedule_interval=daily",
                    "--owner=test@example.org",
                    "--description=test",
                    "--start_date=2020-01-01",
                    "--tag=impact/tier_3",
                ],
            )
            assert result.exit_code == 0

            with open("dags.yaml", "r") as dags_file:
                dags_conf = yaml.safe_load(dags_file.read())
                assert "bqetl_new_dag" in dags_conf
                assert "bqetl_test" in dags_conf
                assert (
                    dags_conf["bqetl_new_dag"]["default_args"]["owner"]
                    == "test@example.org"
                )
                assert dags_conf["bqetl_new_dag"]["description"] == "test"
                assert (
                    dags_conf["bqetl_new_dag"]["default_args"]["start_date"]
                    == "2020-01-01"
                )
                assert dags_conf["bqetl_new_dag"]["schedule_interval"] == "daily"

    def test_dag_remove_non_existing_dag(self, runner):
        with runner.isolated_filesystem():
            dags_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                },
                "bqetl_test_2": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                },
            }

            os.makedirs("sql/moz-fx-data-shared-prod")
            os.mkdir("dags")

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dags_conf))

            result = runner.invoke(
                remove,
                ["non_existing_dag"],
            )
            assert result.exit_code == 1

    def test_dag_remove(self, runner):
        with runner.isolated_filesystem():
            dags_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                },
                "bqetl_test_2": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                },
            }

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dags_conf))

            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql",
                "w",
            ) as f:
                f.write("SELECT 1")

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            os.mkdir("dags")
            with open("dags/bqetl_test.py", "w") as f:
                f.write("")

            result = runner.invoke(
                remove,
                ["bqetl_test"],
            )
            assert result.exit_code == 0
            assert os.listdir("dags") == []

            with open("dags.yaml", "r") as dags_file:
                dags_conf = yaml.safe_load(dags_file.read())
                assert "bqetl_test_2" in dags_conf
                assert "bqetl_test" not in dags_conf

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "r",
            ) as f:
                metadata = yaml.safe_load(f.read())
                assert "scheduling" not in metadata

    def test_dag_generate_without_any_tasks(self, runner):
        with runner.isolated_filesystem():
            dags_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                },
            }

            os.makedirs("sql/moz-fx-data-shared-prod")
            os.mkdir("dags")

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dags_conf))

            result = runner.invoke(
                generate,
                ["bqetl_test"],
            )

            assert result.exit_code == 0



from datetime import date

import pytest

from bigquery_etl.backfill.date_range import BackfillDateRange, get_backfill_partition
from bigquery_etl.metadata.parse_metadata import PartitionType


class TestBackfillDateRange:
    def test_basic_date_range(self):

        date_range = BackfillDateRange(
            start_date=date.fromisoformat("2020-01-01"),
            end_date=date.fromisoformat("2020-01-05"),
        )
        dates = list(date_range)
        assert dates == [
            date.fromisoformat("2020-01-01"),
            date.fromisoformat("2020-01-02"),
            date.fromisoformat("2020-01-03"),
            date.fromisoformat("2020-01-04"),
            date.fromisoformat("2020-01-05"),
        ]

    def test_excludes(self):
        date_range = BackfillDateRange(
            start_date=date.fromisoformat("2020-01-01"),
            end_date=date.fromisoformat("2020-01-05"),
            excludes=[
                date.fromisoformat("2020-01-02"),
                date.fromisoformat("2020-01-03"),
            ],
        )
        dates = list(date_range)

        assert dates == [
            date.fromisoformat("2020-01-01"),
            date.fromisoformat("2020-01-04"),
            date.fromisoformat("2020-01-05"),
        ]

    def test_monthly(self):
        date_range = BackfillDateRange(
            start_date=date.fromisoformat("2020-01-01"),
            end_date=date.fromisoformat("2020-03-05"),
            excludes=[date.fromisoformat("2020-02-02")],
            range_type=PartitionType.MONTH,
        )
        dates = list(date_range)

        assert dates == [
            date.fromisoformat("2020-01-01"),
            date.fromisoformat("2020-03-01"),
        ]


def test_get_backfill_partition():

    assert (
        get_backfill_partition(
            backfill_date=date.fromisoformat("2020-01-03"),
            date_partition_parameter="submission_date",
            date_partition_offset=-1,
            partitioning_type=PartitionType.DAY,
        )
        == "20200102"
    )

    assert (
        get_backfill_partition(
            backfill_date=date.fromisoformat("2020-01-01"),
            date_partition_parameter="submission_date",
            date_partition_offset=0,
            partitioning_type=PartitionType.MONTH,
        )
        == "202001"
    )

    with pytest.raises(NotImplementedError):
        get_backfill_partition(
            backfill_date=date.fromisoformat("2020-01-01"),
            date_partition_parameter="submission_date",
            date_partition_offset=-1,
            partitioning_type=PartitionType.MONTH,
        )



"""Tools for working with date ranges related to running Backfills."""

import logging
from dataclasses import dataclass, field
from datetime import date, timedelta
from typing import Optional

from dateutil.rrule import MONTHLY, rrule

from ..metadata.parse_metadata import PartitionType

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


@dataclass
class BackfillDateRange:
    """An iterator for dates in a range for backfills."""

    start_date: date
    end_date: date
    excludes: Optional[list[date]] = None
    range_type: PartitionType = PartitionType.DAY
    _dates: list[date] = field(init=False)
    _date_index: int = field(init=False)

    def __post_init__(self):
        """Initialize internal attributes."""
        if self.range_type not in (PartitionType.DAY, PartitionType.MONTH):
            raise ValueError(f"Unsupported partitioning type: {self.range_type}")

        match self.range_type:
            case PartitionType.DAY:
                self._dates = [
                    self.start_date + timedelta(i)
                    for i in range((self.end_date - self.start_date).days + 1)
                ]
            case PartitionType.MONTH:
                self._dates = [
                    dt.date()
                    for dt in rrule(
                        freq=MONTHLY,
                        dtstart=self.start_date.replace(day=1),
                        until=self.end_date,
                    )
                ]
                # Dates in excluded must be the first day of the month to match `dates`
                log.info("Converting excludes to first day of the month.")
                if self.excludes is not None:
                    self.excludes = [day.replace(day=1) for day in self.excludes]

    def __iter__(self):
        """Make iterator object."""
        self._date_index = 0
        return self

    def __next__(self):
        """Get the next date in the iteration."""
        if self._date_index < len(self._dates):
            _date = self._dates[self._date_index]
            self._date_index += 1
            if self._is_date_excluded(_date):
                log.info(f"Skipping excluded date: {_date}")
                return self.__next__()
            else:
                return _date
        else:
            raise StopIteration

    def _is_date_excluded(self, d: date) -> bool:
        return self.excludes is not None and d in self.excludes


def _get_offset_date(d: date, date_type: PartitionType, offset: int) -> date:
    match date_type:
        case PartitionType.DAY:
            return d + timedelta(days=offset)
        case PartitionType.MONTH:
            if offset != 0:
                raise NotImplementedError("Offsets unsupported for monthly partitions.")
            return d
        case _:
            raise NotImplementedError(f"Unsupported date type {date_type}.")


def get_backfill_partition(
    backfill_date: date,
    date_partition_parameter: Optional[str],
    date_partition_offset: int,
    partitioning_type: PartitionType,
) -> Optional[str]:
    """Get the partition to write to for a particular backfill date based on query settings."""
    # If date_partition_parameter is explicitly set to None the target is the entire table
    # However if there's a date_partition_offset, the target is still a specific partition
    if date_partition_parameter is None and date_partition_offset == 0:
        return None

    partition_date = _get_offset_date(
        backfill_date, partitioning_type, date_partition_offset
    )

    partition = None
    match partitioning_type:
        case PartitionType.DAY:
            partition = partition_date.strftime("%Y%m%d")
        case PartitionType.MONTH:
            partition = partition_date.strftime("%Y%m")

    return partition



"""Experiment monitoring materialized view generation."""

import os
from pathlib import Path

import click
import yaml
from jinja2 import Environment, FileSystemLoader
from fnmatch import fnmatchcase

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import write_sql

FILE_PATH = Path(os.path.dirname(__file__))
BASE_DIR = Path(FILE_PATH).parent.parent


def generate_queries(project, path, write_dir):
    """Generate experiment monitoring views."""
    with open(Path(path) / "templating.yaml", "r") as f:
        template_config = yaml.safe_load(f) or {}

    for query, args in template_config["queries"].items():
        template_query_dir = FILE_PATH / "templates" / query
        env = Environment(
            loader=FileSystemLoader(template_query_dir),
            keep_trailing_newline=True,
        )
        sql_templates = list(template_query_dir.glob("*.sql"))
        sql_template_file = sql_templates[0].name
        sql_template = env.get_template(sql_template_file)
        metadata_template = env.get_template("metadata.yaml")

        args["destination_table"] = query
        args["search_metrics"] = template_config["search_metrics"]

        if args["per_app"]:
            # generate a separate query for each application dataset
            for dataset in template_config["applications"]:
                if "skip_applications" in args:
                    if any(fnmatchcase(dataset, skip_app) for skip_app in args["skip_applications"]):
                        continue

                args["dataset"] = dataset

                write_sql(
                    write_dir / project,
                    f"{project}.{dataset}_derived.{query}",
                    sql_template_file,
                    reformat(sql_template.render(**args)),
                )

                write_path = Path(write_dir) / project / (dataset + "_derived") / query
                (write_path / "metadata.yaml").write_text(
                    metadata_template.render(**args)
                )
        else:
            # generate a single query that UNIONs application datasets
            # these queries are written to `telemetry`
            args["applications"] = template_config["applications"]

            if "skip_applications" in args:
                args["applications"] = [
                    app
                    for app in args["applications"]
                    if not any(fnmatchcase(app, skip_app) for skip_app in args["skip_applications"])
                ]

            destination_dataset = args.get("destination_dataset", "telemetry_derived")

            write_sql(
                write_dir / project,
                f"{project}.{destination_dataset}.{query}",
                sql_template_file,
                reformat(sql_template.render(**args)),
            )

            write_path = Path(write_dir) / project / destination_dataset / query
            (write_path / "metadata.yaml").write_text(metadata_template.render(**args))


@click.command("generate")
@click.option(
    "--target-project",
    "--target_project",
    help="Which project the queries should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--path",
    help="Where query directories will be searched for.",
    default="sql_generators/experiment_monitoring/templates",
    required=False,
    type=click.Path(file_okay=False),
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@use_cloud_function_option
def generate(target_project, path, output_dir, use_cloud_function):
    """Generate the experiment monitoring views."""
    output_dir = Path(output_dir)
    generate_queries(target_project, path, output_dir)



from pathlib import Path
from textwrap import dedent

import yaml
from google.cloud.bigquery import SchemaField

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.schema import Schema

TEST_DIR = Path(__file__).parent.parent


class TestQuerySchema:
    def test_from_schema_file(self):
        schema_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "schema.yaml"
        )

        schema = Schema.from_schema_file(schema_file)
        assert len(schema.schema["fields"]) == 3

    def test_from_query_file(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_non_incremental_export_v1"
            / "query.sql"
        )

        schema = Schema.from_query_file(query_file)
        assert len(schema.schema["fields"]) == 3

    def test_from_json(self):
        json_schema = {
            "fields": [
                {"name": "amount", "type": "INTEGER"},
                {"name": "amount_captured", "type": "INTEGER"},
            ]
        }

        schema = Schema.from_json(json_schema)
        assert len(schema.schema["fields"]) == 2

    def test_equal_schemas(self):
        empty_schema = Schema.from_json({"fields": {}})
        assert empty_schema.equal(empty_schema)

        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - mode: NULLABLE
              name: client_id
              type: STRING
            - fields:
              - mode: NULLABLE
                name: campaign
                type: STRING
              - mode: NULLABLE
                name: content
                type: STRING
              mode: NULLABLE
              name: attribution
              type: RECORD
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: client_id
              type: STRING
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - fields:
              - mode: NULLABLE
                name: content
                type: STRING
                description: "Cool content"
              - mode: NULLABLE
                name: campaign
                type: STRING
              mode: NULLABLE
              name: attribution
              type: RECORD
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))

        assert schema_1.equal(schema_2) is True
        assert schema_2.equal(schema_1) is True

    def test_schemas_unequal_attributes(self):
        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: INTEGER
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))

        assert schema_1.equal(schema_2) is False
        assert schema_2.equal(schema_1) is False

    def test_schemas_unequal_fields(self):
        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - mode: NULLABLE
              name: client_id
              type: STRING
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: client_id
              type: STRING
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))

        assert schema_1.equal(schema_2) is False
        assert schema_2.equal(schema_1) is False

    def test_schemas_unequal_nested_record(self):
        schema_1_yaml = dedent(
            """
            fields:
            - fields:
              - mode: NULLABLE
                name: multiprocess_compatible
                type: BOOLEAN
            mode: REPEATED
            name: active_addons
            type: RECORD
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - fields:
              - mode: NULLABLE
                name: multiprocess
                type: BOOLEAN
            mode: REPEATED
            name: active_addons
            type: RECORD
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))

        assert schema_1.equal(schema_2) is False
        assert schema_2.equal(schema_1) is False

    def test_schemas_different_descriptions(self):
        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              description: "The submission_date"
              type: DATE
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              description: "Date of the submission"
              type: DATE
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))

        assert schema_1.equal(schema_2) is True
        assert schema_2.equal(schema_1) is True

    def test_schemas_compatible(self):
        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - mode: NULLABLE
              name: client_id
              type: STRING
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))

        assert schema_1.compatible(schema_2) is True
        assert schema_2.compatible(schema_1) is False

    def test_merge_empty_schema(self):
        schema_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - mode: NULLABLE
              name: client_id
              type: STRING
            - fields:
              - mode: NULLABLE
                name: campaign
                type: STRING
              - mode: NULLABLE
                name: content
                type: STRING
              mode: NULLABLE
              name: attribution
              type: RECORD
            """
        )

        schema = Schema.from_json(yaml.safe_load(schema_yaml))
        empty_schema = Schema.from_json({"fields": []})
        schema.merge(empty_schema)
        assert schema.to_json() != empty_schema.to_json()
        assert (
            schema.to_json() == Schema.from_json(yaml.safe_load(schema_yaml)).to_json()
        )

        empty_schema.merge(schema)
        assert (
            empty_schema.to_json()
            == Schema.from_json(yaml.safe_load(schema_yaml)).to_json()
        )

    def test_merge_compatible_schemas(self):
        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - mode: NULLABLE
              name: client_id
              type: STRING
            - fields:
              - mode: NULLABLE
                name: campaign
                type: STRING
              - mode: NULLABLE
                name: content
                type: STRING
              mode: NULLABLE
              name: attribution
              type: RECORD
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - mode: NULLABLE
              name: sample_id
              type: INTEGER
            - fields:
              - mode: NULLABLE
                name: description
                type: STRING
              - mode: NULLABLE
                name: content
                description: "Content description"
                type: STRING
              mode: NULLABLE
              name: attribution
              type: RECORD
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))

        assert len(schema_1.schema["fields"]) == 3
        schema_1.merge(schema_2)
        assert len(schema_1.schema["fields"]) == 4

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        assert len(schema_2.schema["fields"]) == 3
        schema_2.merge(schema_1)
        assert len(schema_2.schema["fields"]) == 4

    def test_merge_different_descriptions(self):
        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              description: "The submission_date"
              type: DATE
            """
        )

        schema_2_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              description: "Date of the submission"
              type: DATE
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_2 = Schema.from_json(yaml.safe_load(schema_2_yaml))
        schema_1.merge(schema_2)

        assert schema_1.schema["fields"][0]["description"] == "The submission_date"

        schema_1_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            """
        )

        schema_1 = Schema.from_json(yaml.safe_load(schema_1_yaml))
        schema_1.merge(schema_2)

        assert schema_1.schema["fields"][0]["description"] == "Date of the submission"

    def test_bigquery_conversion(self):
        bq_schema = [
            SchemaField(
                "record_field",
                "RECORD",
                "REPEATED",
                description="Record field",
                fields=(
                    SchemaField(
                        "nested_field_1",
                        "STRING",
                        "NULLABLE",
                        description="Nested Field 1",
                    ),
                    SchemaField(
                        "nested_field_2",
                        "INTEGER",
                        "NULLABLE",
                        description="Nested Field 2",
                    ),
                ),
            ),
            SchemaField("normal_field", "STRING"),
        ]

        schema = Schema.from_bigquery_schema(bq_schema)
        assert schema.schema["fields"][0]["description"] == "Record field"
        assert schema.schema["fields"][0]["name"] == "record_field"
        assert (
            schema.schema["fields"][0]["fields"][0]["description"] == "Nested Field 1"
        )
        assert schema.schema["fields"][0]["fields"][0]["type"] == "STRING"
        assert schema.schema["fields"][1]["name"] == "normal_field"

        schema_yaml = dedent(
            """
            fields:
            - mode: NULLABLE
              name: submission_date
              type: DATE
            - mode: NULLABLE
              name: client_id
              type: STRING
            - fields:
              - mode: NULLABLE
                name: campaign
                type: STRING
                description: attribution campaign
              mode: NULLABLE
              name: attribution
              type: RECORD
            """
        )

        schema = Schema.from_json(yaml.safe_load(schema_yaml))
        bq_schema = schema.to_bigquery_schema()
        assert len(bq_schema) == 3
        assert bq_schema[0] == SchemaField(
            name="submission_date", field_type="DATE", mode="NULLABLE"
        )
        assert bq_schema[2] == SchemaField(
            name="attribution",
            field_type="RECORD",
            mode="NULLABLE",
            fields=(
                SchemaField(
                    name="campaign",
                    field_type="STRING",
                    description="attribution campaign",
                ),
            ),
        )


def test_generate_compatible_select_expression():
    source_schema = {
        "fields": [
            {"name": "scalar", "type": "INTEGER"},
            {"name": "mismatch_scalar", "type": "INTEGER"},
            {
                "name": "record",
                "type": "RECORD",
                "fields": [
                    {"name": "nested_extra", "type": "DATE"},
                    {
                        "name": "nested_record",
                        "type": "RECORD",
                        "fields": [
                            {"name": "v1", "type": "INTEGER"},
                            {"name": "v2", "type": "INTEGER"},
                        ],
                    },
                    {
                        "name": "mismatch_record",
                        "type": "RECORD",
                        "fields": [{"name": "nested_str", "type": "STRING"}],
                    },
                ],
            },
            {"name": "array_scalar", "type": "INTEGER", "mode": "REPEATED"},
            {
                "name": "array_record",
                "type": "RECORD",
                "mode": "REPEATED",
                "fields": [
                    {"name": "key", "type": "STRING"},
                    {"name": "value", "type": "STRING"},
                ],
            },
            {"name": "extra", "type": "STRING"},
        ]
    }
    target_schema = {
        "fields": [
            {"name": "scalar", "type": "INTEGER"},
            {"name": "mismatch_scalar", "type": "STRING"},
            {
                "name": "record",
                "type": "RECORD",
                "fields": [
                    {"name": "nested_missing", "type": "DATE"},
                    {
                        "name": "nested_record",
                        "type": "RECORD",
                        "fields": [
                            {"name": "v1", "type": "INTEGER"},
                            {"name": "v2", "type": "INTEGER"},
                        ],
                    },
                    {
                        "name": "mismatch_record",
                        "type": "RECORD",
                        "fields": [{"name": "nested_int", "type": "INTEGER"}],
                    },
                ],
            },
            {"name": "array_scalar", "type": "INTEGER", "mode": "REPEATED"},
            {
                "name": "array_record",
                "type": "RECORD",
                "mode": "REPEATED",
                "fields": [
                    {"name": "value", "type": "STRING"},
                    {"name": "key", "type": "STRING"},
                ],
            },
            {"name": "missing", "type": "STRING"},
        ]
    }
    expected_expr = """
    scalar,
    CAST(NULL AS STRING) AS `mismatch_scalar`,
    STRUCT(
        CAST(NULL AS DATE) AS `nested_missing`,
        record.nested_record,
        STRUCT(CAST(NULL AS INTEGER) AS `nested_int`) AS `mismatch_record`
    ) AS `record`,
    array_scalar,
    ARRAY(
        SELECT
            STRUCT(
                array_record.value,
                array_record.key
            )
        FROM
            UNNEST(array_record) AS `array_record`
    ) AS `array_record`,
    CAST(NULL AS STRING) AS `missing`
    """

    source = Schema.from_json(source_schema)
    target = Schema.from_json(target_schema)

    select_expr = source.generate_compatible_select_expression(
        target, unnest_structs=False
    )
    assert reformat(select_expr) == reformat(expected_expr)


def test_generate_select_expression_unnest_struct():
    """unnest_struct argument should unnest records even when they match."""
    source_schema = {
        "fields": [
            {
                "name": "record",
                "type": "RECORD",
                "fields": [
                    {"name": "key", "type": "STRING"},
                    {"name": "value", "type": "STRING"},
                ],
            },
        ]
    }

    source = Schema.from_json(source_schema)

    unnest_expr = source.generate_select_expression(unnest_structs=True)
    assert reformat(unnest_expr) == reformat(
        "STRUCT(record.key, record.value) AS `record`"
    )

    no_unnest_expr = source.generate_select_expression(unnest_structs=False)
    assert reformat(no_unnest_expr) == reformat("record")


def test_generate_select_expression_remove_fields():
    """remove_fields argument remove the given fields from the output."""
    source_schema = {
        "fields": [
            {
                "name": "record",
                "type": "RECORD",
                "fields": [
                    {"name": "key", "type": "STRING"},
                    {"name": "value", "type": "STRING"},
                ],
            },
            {"name": "scalar", "type": "INTEGER"},
        ]
    }

    source = Schema.from_json(source_schema)

    unnest_expr = source.generate_select_expression(
        unnest_structs=True, remove_fields=["record.value", "scalar"]
    )
    assert reformat(unnest_expr) == reformat("STRUCT(record.key) AS `record`")


def test_generate_select_expression_max_unnest_depth():
    """max_unnest_depth argument should stop unnesting at the given depth."""
    source_schema = {
        "fields": [
            {
                "name": "record",
                "type": "RECORD",
                "fields": [
                    {
                        "name": "record2",
                        "type": "RECORD",
                        "fields": [
                            {
                                "name": "record3",
                                "type": "RECORD",
                                "fields": [{"name": "key", "type": "STRING"}],
                            },
                        ],
                    },
                ],
            },
        ]
    }

    source = Schema.from_json(source_schema)

    expected_expr = """
    STRUCT(
        STRUCT(
            record.record2.record3
        ) AS `record2`
    ) AS `record`
    """

    unnest_expr = source.generate_select_expression(
        unnest_structs=True, max_unnest_depth=2
    )
    assert reformat(unnest_expr) == reformat(expected_expr)


def test_generate_select_expression_unnest_allowlist():
    """unnest_allowlist argument should cause only the given fields to be unnested."""
    source_schema = {
        "fields": [
            {
                "name": "record",
                "type": "RECORD",
                "fields": [{"name": "key", "type": "STRING"}],
            },
            {
                "name": "record2",
                "type": "RECORD",
                "fields": [
                    {
                        "name": "record3",
                        "type": "RECORD",
                        "fields": [{"name": "key", "type": "STRING"}],
                    }
                ],
            },
        ]
    }

    source = Schema.from_json(source_schema)

    expected_expr = """
        record,
        STRUCT(
            STRUCT(
                record2.record3.key
            ) AS `record3`
        ) AS `record2`
        """

    unnest_expr = source.generate_select_expression(
        unnest_structs=True, unnest_allowlist=["record2"]
    )
    assert reformat(unnest_expr) == reformat(expected_expr)



import os
from unittest.mock import patch

import pytest
import yaml
from click.testing import CliRunner

from bigquery_etl.cli.static import publish

VALID_WORKGROUP_ACCESS = [
    dict(
        role="roles/bigquery.dataViewer",
        members=["workgroup:mozilla-confidential"],
    )
]

TABLE_METADATA_CONF = {
    "friendly_name": "test",
    "description": "test",
    "owners": ["test@example.org"],
    "workgroup_access": VALID_WORKGROUP_ACCESS,
}

DATASET_METADATA_CONF = {
    "friendly_name": "test",
    "description": "test",
    "dataset_base_acl": "derived",
    "workgroup_access": VALID_WORKGROUP_ACCESS,
}


class TestStatic:

    @pytest.fixture
    def runner(self):
        return CliRunner()

    @patch("google.cloud.bigquery.Client")
    def test_static_publish(self, mock_client, runner):
        with runner.isolated_filesystem():
            SQL_DIR = "sql/moz-fx-data-shared-prod/test/test_data_v1"
            os.makedirs(SQL_DIR)

            with open(
                "sql/moz-fx-data-shared-prod/test/test_data_v1/data.csv", "w"
            ) as f:
                f.write("")

            with open(
                "sql/moz-fx-data-shared-prod/test/test_data_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(TABLE_METADATA_CONF))

            with open(
                "sql/moz-fx-data-shared-prod/test/dataset_metadata.yaml", "w"
            ) as f:
                f.write(yaml.dump(DATASET_METADATA_CONF))

            result = runner.invoke(publish)

            assert result.exit_code == 0
            assert mock_client.return_value.dataset.call_count == 1
            assert mock_client.return_value.dataset.call_args.args[0] == "test"
            assert mock_client.return_value.load_table_from_file.call_count == 1
            assert (
                mock_client.return_value.get_table.call_count == 1
            )  # from publish metadata
            assert (
                mock_client.return_value.update_table.call_count == 1
            )  # from publish metadata



#!/usr/bin/env python3

"""Search for tables with client id columns."""
import datetime
from collections import defaultdict
from functools import cache
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Any, Dict, Iterable, List, Set

import click
from google.cloud import bigquery
from google.cloud import datacatalog_lineage_v1 as datacatalog_lineage
from google.cloud.bigquery import TableReference
from google.cloud.exceptions import NotFound

from bigquery_etl.schema import Schema
from bigquery_etl.shredder.config import (
    CLIENT_ID,
    DELETE_TARGETS,
    GLEAN_CLIENT_ID,
    SHARED_PROD,
    DeleteSource,
    find_glean_targets,
    get_glean_channel_to_app_name_mapping,
)

FIND_TABLES_QUERY_TEMPLATE = """
WITH no_client_id_tables AS (
  SELECT
    table_catalog,
    table_schema,
    table_name,
  FROM
    `{project}.region-us.INFORMATION_SCHEMA.TABLE_OPTIONS`
  WHERE
    option_name = 'labels'
    AND option_value LIKE '%(\"include_client_id\", \"false\")%'
)

SELECT
  table_catalog,
  table_schema,
  table_name,
FROM
  `{project}.region-us.INFORMATION_SCHEMA.TABLES`
LEFT JOIN
  no_client_id_tables
USING
  (table_catalog, table_schema, table_name)
WHERE
  -- find tables with columns ending with client_id
  (
    (
      ddl LIKE '%client_id STRING%' AND no_client_id_tables.table_name IS NULL
    )
    OR (
      -- glean tables may have an all null client_info.client_id column but may have a secondary client id
      ddl LIKE '%client_id STRING%client_id STRING%' AND no_client_id_tables.table_name IS NOT NULL
    )
  )
  AND table_type = 'BASE TABLE'  -- exclude views
  AND (
    table_schema LIKE '%_derived'
    or table_schema LIKE '%_stable'
  )
  -- TODO: can't get lineage for most opmon tables, need to figure this out separately
  AND table_schema != "operational_monitoring_derived"
  AND table_schema != "backfills_staging_derived"
  AND table_name != 'deletion_request_v1'
  AND table_name != 'deletion_request_v4'
"""


def find_client_id_tables(project: str) -> List[str]:
    """Return a list of tables that have columns ending with 'client_id'."""
    client = bigquery.Client(project=project)
    row_results = client.query_and_wait(
        query=FIND_TABLES_QUERY_TEMPLATE.format(project=project)
    )

    return [f"{project}.{row.table_schema}.{row.table_name}" for row in row_results]


def get_upstream_stable_tables(id_tables: List[str]) -> Dict[str, Set[str]]:
    """Build map of tables to upstream stable tables using GCP data catalog lineage.

    Note that the data catalog only uses the information from the last 30 days of query jobs.
    """
    client = datacatalog_lineage.LineageClient()

    upstream_stable_tables = defaultdict(set)

    def traverse_upstream(base_table: str):
        """Recursively traverse lineage to find stable tables."""
        table_ref = TableReference.from_string(base_table)

        if table_ref.dataset_id.endswith("_stable"):  # stable tables are terminal nodes
            upstream_stable_tables[base_table] = {base_table}
        elif base_table not in upstream_stable_tables:
            upstream_links_result = client.search_links(
                request={
                    "parent": "projects/moz-fx-data-shared-prod/locations/us",
                    "target": datacatalog_lineage.EntityReference(
                        fully_qualified_name=f"bigquery:{base_table}"
                    ),
                }
            )
            # recursively add upstream tables
            for upstream_link in upstream_links_result:
                # remove "bigquery:" and "sharded:" prefixes
                parent_table = upstream_link.source.fully_qualified_name.split(":")[-1]

                upstream_stable_tables[base_table] = upstream_stable_tables[
                    base_table
                ].union(traverse_upstream(parent_table))

        return upstream_stable_tables[base_table]

    upstream_stable_table_map = {}

    print("Upstream stable tables:")
    for table_name in id_tables:
        upstream_stable_table_map[table_name] = set(traverse_upstream(table_name))
        print(f"{table_name} upstream: {upstream_stable_table_map[table_name]}")

    return upstream_stable_table_map


@cache
def table_exists(client: bigquery.Client, table_name: str) -> bool:
    """Return true if given project.dataset.table exists, caching results."""
    try:
        client.get_table(table_name)
        return True
    except NotFound:
        return False


def get_associated_deletions(
    project: str,
    upstream_stable_tables: Dict[str, Set[str]]
) -> Dict[str, Set[DeleteSource]]:
    """Get a list of associated deletion requests tables per table based on the stable tables."""
    client = bigquery.Client(project=project)

    # deletion targets for stable tables defined in the shredder config
    known_stable_table_sources: Dict[str, Set[DeleteSource]] = {
        f"{target.project}.{target.dataset_id}.{target.table_id}": (
            set(src) if isinstance(src, Iterable) else {src}
        )
        for target, src in DELETE_TARGETS.items()
        if target.dataset_id.endswith("_stable")
    }

    table_to_deletions: Dict[str, Set[DeleteSource]] = {}

    datasets_with_additional_deletion_requests = {}

    for base_table in upstream_stable_tables:
        if base_table.endswith(".additional_deletion_requests_v1"):
            dataset_name = TableReference.from_string(base_table).dataset_id
            datasets_with_additional_deletion_requests[dataset_name] = (
                f"{dataset_name}.additional_deletion_requests_v1"
            )
            datasets_with_additional_deletion_requests[
                dataset_name.replace("_derived", "_stable")
            ] = f"{dataset_name}.additional_deletion_requests_v1"

    glean_channel_names = get_glean_channel_to_app_name_mapping()

    for table_name, stable_tables in upstream_stable_tables.items():
        deletion_tables: Set[DeleteSource] = set()

        for stable_table in stable_tables:
            if stable_table in known_stable_table_sources:
                table_to_deletions[stable_table] = known_stable_table_sources[
                    stable_table
                ]
            elif stable_table not in table_to_deletions:
                stable_table_ref = TableReference.from_string(stable_table)

                # glean table
                if (
                    stable_table_ref.dataset_id[: -len("_stable")]
                    in glean_channel_names
                ):
                    if table_exists(
                        client,
                        f"{stable_table_ref.project}.{stable_table_ref.dataset_id}.deletion_request_v1",
                    ):
                        table_to_deletions[stable_table] = {
                            DeleteSource(
                                table=f"{stable_table_ref.dataset_id}.deletion_request_v1",
                                field=GLEAN_CLIENT_ID,
                                project=SHARED_PROD,
                            )
                        }
                    else:
                        print(f"No deletion requests for {stable_table_ref.dataset_id}")
                        table_to_deletions[stable_table] = set()

                    if (
                        stable_table_ref.dataset_id
                        in datasets_with_additional_deletion_requests
                    ):
                        table_to_deletions[stable_table].add(
                            DeleteSource(
                                table=datasets_with_additional_deletion_requests[
                                    stable_table_ref.dataset_id
                                ],
                                field=CLIENT_ID,
                                project=SHARED_PROD,
                            )
                        )
                # unknown legacy telemetry or non-glean structured
                else:
                    table_to_deletions[stable_table] = set()

            deletion_tables = deletion_tables.union(table_to_deletions[stable_table])

        table_to_deletions[table_name] = deletion_tables

    return {
        table: deletions
        for table, deletions in table_to_deletions.items()
        if table in upstream_stable_tables
    }


def delete_source_to_dict(source: DeleteSource):
    """Convert a DeleteSource to a dict, removing the condition field."""
    d = source.__dict__.copy()
    d.pop("conditions")
    return d


def get_missing_deletions(
    associated_deletions: Dict[str, Set[DeleteSource]]
) -> List[Dict[str, Any]]:
    """Get list of all tables with the currently configured deletion sources and the sources based on lineage."""
    # get the generated glean deletion list
    with ThreadPool(processes=12) as pool:
        bigquery_client = bigquery.Client()
        glean_delete_targets = find_glean_targets(pool, client=bigquery_client)

    glean_channel_names = get_glean_channel_to_app_name_mapping()
    glean_app_name_to_channels = defaultdict(list)
    for channel, app_name in glean_channel_names.items():
        glean_app_name_to_channels[app_name].append(channel)

    table_deletions = []

    for target, sources in (*glean_delete_targets.items(), *DELETE_TARGETS.items()):
        target_table = f"{target.project}.{target.table}"

        # expand per-app deletion request views into per-channel tables
        unnested_sources = set()
        if isinstance(sources, Iterable):
            for source in sources:
                if (
                    source.dataset_id in glean_app_name_to_channels
                    and source.table_id == "deletion_request"
                ):
                    for channel in glean_app_name_to_channels[source.dataset_id]:
                        unnested_sources.add(
                            DeleteSource(
                                table=f"{channel}_stable.deletion_request_v1",
                                field=source.field,
                                project=source.project,
                                conditions=source.conditions,
                            )
                        )
                else:
                    unnested_sources.add(source)
        else:
            unnested_sources.add(sources)

        # tables not in associated_deletions likely use another id column, e.g. user_id
        detected_deletions = associated_deletions.pop(target_table, set())

        table_deletions.append(
            {
                "project_id": target.project,
                "dataset_id": target.dataset_id,
                "table_id": target.table_id,
                "current_sources": [
                    delete_source_to_dict(source) for source in unnested_sources
                ],
                "detected_sources": [
                    delete_source_to_dict(detected) for detected in detected_deletions
                ],
                "matching_sources": set(unnested_sources) == detected_deletions,
            }
        )

    # delete target not in shredder config
    for table_name, srcs in associated_deletions.items():
        project, dataset, table = table_name.split(".")
        table_deletions.append(
            {
                "project_id": project,
                "dataset_id": dataset,
                "table_id": table,
                "current_sources": [],
                "detected_sources": [
                    delete_source_to_dict(detected) for detected in srcs
                ],
                "matching_sources": len(srcs) == 0,
            }
        )

    return table_deletions


def write_to_bigquery(
    run_date: datetime.datetime,
    target_table: TableReference,
    deletions: List[Dict[str, Any]],
):
    client = bigquery.Client()

    result = client.load_table_from_json(
        json_rows=[
            {"run_date": run_date.date().isoformat(), **deletion}
            for deletion in deletions
        ],
        destination=f"{str(target_table)}${run_date.date().strftime('%Y%m%d')}",
        job_config=bigquery.LoadJobConfig(
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
            schema=Schema.from_schema_file(
                Path(__file__).parent / "schema.yaml"
            ).to_bigquery_schema(),
            time_partitioning=bigquery.TimePartitioning(field="run_date"),
        ),
    ).result()

    print(f"Wrote {result.output_rows} rows to {result.destination}")


@click.command
@click.option(
    "--run-date", type=click.DateTime(), help="The date to write in the output."
)
@click.option(
    "--output-table",
    type=TableReference.from_string,
    metavar="PROJECT.DATASET.TABLE",
    help="Table to write results to in the form of PROJECT.DATASET.TABLE.",
)
@click.option(
    "--project-id",
    default=SHARED_PROD,
    help="BigQuery project to search for client id tables.",
)
def main(run_date, output_table, project_id):
    """Find tables in the given project that could be added to shredder."""
    # TODO: handle other id columns
    client_id_tables = find_client_id_tables(project_id)

    print(f"Found {len(client_id_tables)} client id tables.")

    upstream_stable_tables = get_upstream_stable_tables(client_id_tables)

    associated_deletions = get_associated_deletions(project_id, upstream_stable_tables)

    table_deletions = get_missing_deletions(associated_deletions)

    write_to_bigquery(run_date, output_table, table_deletions)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Monthly data exports of MDN 'Popularities'. This aggregates and counts total page visits."""
import logging
from argparse import ArgumentParser
from datetime import datetime
from os.path import join
from uuid import uuid4

from google.cloud import bigquery, storage

QUERY_TEMPLATE = """\
SELECT REGEXP_EXTRACT(pge.metrics.url2.page_path, r'^https://developer.mozilla.org(/.+?/docs/[^?#]+)') AS Page,
       COUNT(*) AS Pageviews,
FROM `moz-fx-data-shared-prod.mdn_yari.page` AS pge
WHERE DATE(submission_timestamp) BETWEEN DATE_TRUNC(@submission_date, MONTH) AND LAST_DAY(@submission_date)
  AND client_info.app_channel = 'prod'
  AND REGEXP_CONTAINS(pge.metrics.url2.page_path, r'^https://developer.mozilla.org(/.+?/docs/[^?#]+)')
  AND pge.metrics.string.page_http_status = '200'
GROUP BY Page
ORDER BY Pageviews DESC
"""

APP = "mdn_yari"

CURRENT_FILE_NAME = "current.csv"

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--date", type=lambda x: datetime.strptime(x, "%Y-%m-%d").date(), required=True
)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--temp_dataset", default="tmp")
parser.add_argument("--temp_table", default="mdn_popularities_v1")
parser.add_argument("--destination_project", default="moz-fx-mdn-prod")
parser.add_argument("--destination_bucket", default="popularities-prod-mdn")
parser.add_argument("--destination_path", default="")


def main():
    """Generate a tmp table and extracts to GCS."""
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    # create temp table
    temp_table = f"{args.project}.{args.temp_dataset}.{args.temp_table}"
    logging.info("Creating a temp table: %s" % temp_table)

    client.query(
        QUERY_TEMPLATE.format(date=args.date),
        job_config=bigquery.QueryJobConfig(
            destination=temp_table,
            write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
            query_parameters=[
                bigquery.ScalarQueryParameter(
                    "submission_date", "DATE", str(args.date)
                ),
            ],
        ),
    ).result()

    # table to GCS object
    dataset_ref = bigquery.DatasetReference(args.project, args.temp_dataset)
    table_ref = dataset_ref.table(args.temp_table)

    target_file_name = f"{uuid4()}.csv"
    target_file_path = join(
        args.destination_path, args.date.strftime("%Y/%m"), target_file_name
    )
    mdn_uri = f"gs://{args.destination_bucket}/{target_file_path}"

    logging.info(
        "Exporting %s to GCS: %s:%s" % (temp_table, args.destination_project, mdn_uri)
    )
    extract_job = client.extract_table(
        table_ref,
        mdn_uri,
        location="US",
    )
    extract_job.result()  # Waits for job to complete.

    logging.info("Deleting temp table: %s" % temp_table)
    client.delete_table(temp_table)

    # Make it available as current.
    current_file_path = join(args.destination_path, CURRENT_FILE_NAME)

    storage_client = storage.Client(args.project)
    bucket = storage_client.get_bucket(args.destination_bucket)
    blob = bucket.get_blob(target_file_path)
    bucket.copy_blob(blob, bucket, current_file_path)


if __name__ == "__main__":
    main()



"""bigquery-etl CLI."""

import logging
import warnings
from pathlib import Path

import rich_click as click

from .._version import __version__

# We rename the import, otherwise it affects monkeypatching in tests
from ..cli.alchemer import alchemer as alchemer_
from ..cli.backfill import backfill
from ..cli.check import check
from ..cli.dag import dag
from ..cli.dryrun import dryrun
from ..cli.format import format
from ..cli.generate import generate
from ..cli.metadata import metadata
from ..cli.monitoring import monitoring
from ..cli.query import query
from ..cli.routine import mozfun, routine
from ..cli.stage import stage
from ..cli.static import static_
from ..cli.view import view
from ..config import ConfigLoader
from ..copy_deduplicate import copy_deduplicate
from ..dependency import dependency
from ..docs import docs_
from ..glam.cli import glam
from ..stripe import stripe_
from ..subplat.apple import apple


def cli(prog_name=None):
    """Create the bigquery-etl CLI."""
    commands = {
        "query": query,
        "dag": dag,
        "dependency": dependency,
        "dryrun": dryrun,
        "generate": generate,
        "format": format,
        "routine": routine,
        "mozfun": mozfun,
        "stripe": stripe_,
        "glam": glam,
        "view": view,
        "alchemer": alchemer_,
        "apple": apple,
        "static": static_,
        "docs": docs_,
        "copy_deduplicate": copy_deduplicate,
        "stage": stage,
        "backfill": backfill,
        "check": check,
        "metadata": metadata,
        "monitoring": monitoring,
    }

    @click.group(commands=commands)
    @click.version_option(version=__version__)
    @click.option(
        "--log-level",
        "--log_level",
        help="Log level.",
        default=logging.getLevelName(logging.INFO),
        type=str.upper,
    )
    def group(log_level):
        """CLI tools for working with bigquery-etl."""
        logging.root.setLevel(level=log_level)

    warnings.filterwarnings(
        "ignore", "Your application has authenticated using end user credentials"
    )

    group(prog_name=prog_name)


if __name__ == "__main__":
    ConfigLoader.set_project_dir(Path().absolute())
    cli()



"""Generate events stream queries for Glean apps."""

import re

from bigquery_etl.config import ConfigLoader
from sql_generators.glean_usage.common import GleanTable, ping_has_metrics

TARGET_TABLE_ID = "events_stream_v1"
PREFIX = "events_stream"


class EventsStreamTable(GleanTable):
    """Represents generated events_stream table."""

    def __init__(self):
        """Initialize events_stream table."""
        GleanTable.__init__(self)
        self.target_table_id = TARGET_TABLE_ID
        self.prefix = PREFIX
        self.per_app_enabled = True
        self.per_app_id_enabled = True
        self.across_apps_enabled = True
        self.cross_channel_template = "cross_channel_events_stream.query.sql"
        self.base_table_name = "events_v1"
        self.custom_render_kwargs = {}

    def generate_per_app_id(
        self,
        project_id,
        baseline_table,
        output_dir=None,
        use_cloud_function=True,
        app_info=[],
        parallelism=8,
        id_token=None,
    ):
        # Get the app ID from the baseline_table name.
        # This is what `common.py` also does.
        app_id = re.sub(r"_stable\..+", "", baseline_table)
        app_id = ".".join(app_id.split(".")[1:])

        # Skip any not-allowed app.
        if app_id in ConfigLoader.get(
            "generate", "glean_usage", "events_stream", "skip_apps", fallback=[]
        ):
            return

        metrics_as_struct = app_id in ConfigLoader.get(
            "generate", "glean_usage", "events_stream", "metrics_as_struct", fallback=[]
        )

        # Separate apps with legacy telemetry client ID vs those that don't have it
        if app_id == "firefox_desktop":
            has_legacy_telemetry_client_id = True
        else:
            has_legacy_telemetry_client_id = False

        # Separate apps with profile group ID vs those that don't have it
        if app_id == "firefox_desktop":
            has_profile_group_id = True
        else:
            has_profile_group_id = False

        unversioned_table_name = re.sub(r"_v[0-9]+$", "", baseline_table.split(".")[-1])

        self.custom_render_kwargs = {
            "has_profile_group_id": has_profile_group_id,
            "has_legacy_telemetry_client_id": has_legacy_telemetry_client_id,
            "metrics_as_struct": metrics_as_struct,
            "has_metrics": ping_has_metrics(app_id, unversioned_table_name),
        }

        super().generate_per_app_id(
            project_id,
            baseline_table,
            output_dir,
            use_cloud_function,
            app_info,
            id_token=id_token,
        )

    def generate_per_app(
        self,
        project_id,
        app_info,
        output_dir=None,
        use_cloud_function=True,
        parallelism=8,
        id_token=None,
    ):
        """Generate the events_stream table query per app_name."""
        target_dataset = app_info[0]["app_name"]
        if target_dataset in ConfigLoader.get(
            "generate", "glean_usage", "events_stream", "skip_datasets", fallback=[]
        ):
            return

        super().generate_per_app(project_id, app_info, output_dir, id_token=id_token)



"""Run a query on clients_scalar_aggregates on one app_version at a time."""

import datetime
import sys
import time
from pathlib import Path

import click
from google.cloud import bigquery

VERSION_QUERY_TEMPLATE = """
SELECT
  DISTINCT(app_version)
FROM
  `{project}.{dataset}.clients_scalar_aggregates_v1`
WHERE
  submission_date = '{date}'
"""

SQL_BASE_DIR = (
    Path(__file__).parent.parent.parent
    / "sql"
    / "moz-fx-data-shared-prod"
    / "telemetry_derived"
)


def default_billing_project(ctx, param, value):
    """Falls back to --project."""
    if value is not None:
        return value
    return ctx.params.get("project", "default_value")


@click.command()
@click.option("--submission-date", required=True, type=datetime.date.fromisoformat)
@click.option("--dst-table", required=True)
@click.option(
    "--project", default="moz-fx-data-shared-prod", help="Project for final table"
)
@click.option(
    "--billing-project", callback=default_billing_project, help="Billing project"
)
@click.option("--tmp-project", help="Project for temporary intermediate tables")
@click.option("--dataset", default="telemetry_derived")
def main(submission_date, dst_table, project, billing_project, tmp_project, dataset):
    """Run query per app_version."""
    bq_client = bigquery.Client(project=billing_project)

    app_versions = [
        row["app_version"]
        for row in bq_client.query(
            VERSION_QUERY_TEMPLATE.format(
                date=submission_date, project=project, dataset=dataset
            )
        ).result()
    ]

    print(f"Found versions: {app_versions}")

    if len(app_versions) == 0:
        print("Source table empty", file=sys.stderr)
        sys.exit(1)

    sql_path = SQL_BASE_DIR / dst_table / "query.sql"

    query_text = sql_path.read_text()

    # Write to intermediate table to avoid partial writes to destination table
    if tmp_project is None:
        tmp_project = project
    intermediate_table = f"{tmp_project}.analysis.glam_temp_clustered_query_{dst_table}"
    print(f"Writing results to {intermediate_table}")

    for i, app_version in enumerate(app_versions):
        print(f"Querying for app_version {app_version}")

        query_config = bigquery.QueryJobConfig(
            query_parameters=[
                bigquery.ScalarQueryParameter(
                    "submission_date", "DATE", str(submission_date)
                ),
                bigquery.ScalarQueryParameter("app_version", "INT64", app_version),
            ],
            clustering_fields=["metric", "channel"],
            destination=intermediate_table,
            default_dataset=f"{project}.{dataset}",
            write_disposition=(
                bigquery.WriteDisposition.WRITE_TRUNCATE
                if i == 0
                else bigquery.WriteDisposition.WRITE_APPEND
            ),
        )

        query_job = bq_client.query(query_text, job_config=query_config)

        # Periodically print so airflow gke operator doesn't think task is dead
        elapsed = 0
        while not query_job.done():
            time.sleep(10)
            elapsed += 10
            if elapsed % 200 == 10:
                print("Waiting on query...")

        print(f"Total elapsed: approximately {elapsed} seconds")

        results = query_job.result()
        print(f"Query job {query_job.job_id} finished")
        print(f"{results.total_rows} rows in {intermediate_table}")

    copy_config = bigquery.CopyJobConfig(
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
    )

    print(f"Copying {intermediate_table} to {project}.{dataset}.{dst_table}")
    bq_client.copy_table(
        intermediate_table,
        f"{project}.{dataset}.{dst_table}",
        job_config=copy_config,
    ).result()

    print(f"Deleting {intermediate_table}")
    bq_client.delete_table(intermediate_table)


if __name__ == "__main__":
    main()



"""Code for setting up the the tests."""

import math
import shutil
from datetime import datetime, timedelta
from pathlib import Path

import yaml

ROOT = Path(__file__).parent

EPOCH = datetime(2014, 12, 28)
START_DATE = datetime(2020, 6, 1)

HISTORY_DAYS = 14
# testing when we hit Firefox 100, soonish
START_VERSION = 100 + HISTORY_DAYS // 2


def app_build(start_date):
    """Generate an app build id per Fenix specifications."""
    return f"{(start_date-EPOCH).days*24<<3:010d}"


def input_row(submission_offset, build_offset, version_offset=0):
    """Generate a single row in a glean metrics ping that is being referenced."""
    return {
        "submission_timestamp": (START_DATE - timedelta(submission_offset)).isoformat(),
        "client_info": {"app_build": app_build(START_DATE - timedelta(build_offset))},
        "metrics": {
            "string": {
                "gecko_version": f"{START_VERSION-version_offset}.0.0",
                "geckoview_version": f"{START_VERSION-version_offset}.0.0",
            },
        },
    }


def main(test_name):
    """Generate assets for each test."""
    assert (ROOT / test_name).is_dir(), f"{ROOT / test_name} not name of test"
    # We use the deprecated nightly product since it is not being filtered
    with (ROOT / test_name / "org_mozilla_fenix_nightly.metrics.yaml").open("w") as fp:
        # Needs some data out of the period for the cold start of the window
        # function. We'll also include dates in the future. There is a new
        # version every day.
        rows = [input_row(i, i, i) for i in range(-10, HISTORY_DAYS + 2)]
        yaml.dump(sorted(rows, key=lambda x: x["client_info"]["app_build"]) * 6, fp)
    # bad rows, versions less than 100 put before and after the 100 mark. The
    # one for fenix will probably get filtered out because of the channel norm
    # udf.
    bad_rows = [
        input_row(HISTORY_DAYS // 2 + 2, 30, 1337),
        input_row(HISTORY_DAYS // 2 - 2, 30, 1337),
    ]

    with (ROOT / test_name / "org_mozilla_fenix.metrics.yaml").open("w") as fp:
        yaml.dump(bad_rows, fp)
    with (ROOT / test_name / "org_mozilla_fennec_aurora.metrics.yaml").open("w") as fp:
        yaml.dump(bad_rows, fp)

    for dataset in [
        "org_mozilla_fenix_nightly",
        "org_mozilla_fenix",
        "org_mozilla_fennec_aurora",
    ]:
        shutil.copyfile(
            ROOT / "metrics.schema.json",
            ROOT / test_name / f"{dataset}.metrics.schema.json",
        )

    init_rows = []
    for hour in range(HISTORY_DAYS * 24 + 1):
        # new version every day
        version_offset = math.ceil(hour / 24)
        init_rows += [
            {
                "build_hour": (START_DATE - timedelta(hours=hour)).isoformat(),
                "geckoview_major_version": START_VERSION - version_offset,
                "n_pings": 6,
            }
        ]
    with (ROOT / test_name / "expect.yaml").open("w") as fp:
        yaml.dump(init_rows, fp)


if __name__ == "__main__":
    main("test_aggregation")



"""Helper methods to fetch BigQuery tables."""

from fnmatch import fnmatchcase
from typing import List

from google.cloud import bigquery


def _uses_wildcards(pattern: str) -> bool:
    return bool(set("*?[]") & set(pattern))


def get_tables_matching_patterns(
    client: bigquery.Client, patterns: List[str]
) -> List[str]:
    """Get BigQuery tables matching the provided patterns."""
    all_projects = None
    all_datasets = {}
    all_tables = {}
    matching_tables = []

    for pattern in patterns:
        project, _, dataset_table = pattern.partition(":")
        dataset, _, table = dataset_table.partition(".")
        projects = [project or client.project]
        dataset = dataset or "*"
        table = table or "*"
        if _uses_wildcards(project):
            if all_projects is None:
                all_projects = [p.project_id for p in client.list_projects()]
            projects = [p for p in all_projects if fnmatchcase(project, p)]
        for project in projects:
            datasets = [dataset]
            if _uses_wildcards(dataset):
                if project not in all_datasets:
                    all_datasets[project] = [
                        d.dataset_id for d in client.list_datasets(project)
                    ]
                datasets = [d for d in all_datasets[project] if fnmatchcase(d, dataset)]
            for dataset in datasets:
                dataset = f"{project}.{dataset}"
                tables = [f"{dataset}.{table}"]
                if _uses_wildcards(table):
                    if dataset not in all_tables:
                        all_tables[dataset] = list(client.list_tables(dataset))
                    tables = [
                        f"{dataset}.{t.table_id}"
                        for t in all_tables[dataset]
                        if fnmatchcase(t.table_id, table)
                    ]
                matching_tables += tables

    return matching_tables



import os
import sys
from pathlib import Path

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from textwrap import dedent

from __init__ import generate_funnels

BASE_DIR = Path(os.path.dirname(__file__)).parent


class TestFunnels:
    def test_generate_funnels(self, tmp_path):
        output_dir = tmp_path / "sql" / "test-project" / "mozilla_vpn_derived"
        output_dir.mkdir(parents=True)

        config_dir = tmp_path / "configs"
        config_dir.mkdir(parents=True)

        config = dedent(
            """
            destination_dataset = "mozilla_vpn_derived"
            platform = "mozilla_vpn"
            owners = ["example@mozilla.org"]  # optional; users getting notification if funnel run fails
            version = "1"  # optional; default is set to 1

            [funnels]

            [funnels.subscription_funnel]

            friendly_name = "Start Subscription Funnel"
            description = "Funnel from Signup to starting a subscription"
            steps = ["signup", "verify", "start_subscription"]
            dimensions = ["os"]

            [funnels.subscription_error_funnel]
            friendly_name = "Subscription Error Funnel"
            description = "Funnel from Signup to running into an error"
            steps = ["signup", "verify", "error_subscription"]


            [steps]

            [steps.signup]
            friendly_name = "Sign up"
            description = "Sign up for VPN"
            data_source = "events"
            filter_expression = '''
                event_name = 'authentication_inapp_step' AND
                `mozfun.map.get_key`(event_extra, 'state') = 'StateVerifyingSessionEmailCode'
            '''
            join_previous_step_on = "client_info.client_id"
            select_expression = "client_info.client_id"
            aggregation = "count distinct"

            [steps.verify]
            friendly_name = "Verify"
            description = "Verify email"
            data_source = "events"
            select_expression = "client_info.client_id"
            where_expression = '''
                event_name = 'authentication_inapp_step' AND
                `mozfun.map.get_key`(event_extra, 'state') = 'StateVerifyingSessionEmailCode'
            '''
            aggregation = "count distinct"
            join_previous_step_on = "client_info.client_id"

            [steps.start_subscription]
            friendly_name = "Start Subscription"
            description = "Start VPN subscription"
            data_source = "events"
            select_expression = "client_info.client_id"
            where_expression = "event_name = 'iap_subscription_started'"
            aggregation = "count distinct"
            join_previous_step_on = "client_info.client_id"

            [steps.error_subscription]
            friendly_name = "Subscription Error"
            description = "subscription error"
            data_source = "events"
            select_expression = "client_info.client_id"
            where_expression = "event_name = 'error_alert_shown'"
            aggregation = "count"
            join_previous_step_on = "client_info.client_id"


            [data_sources]

            [data_sources.events]
            from_expression = '''
                (SELECT * FROM mozdata.mozilla_vpn.events_unnested
                WHERE client_info.app_channel = 'production' AND client_info.os = 'iOS')
            '''
            submission_date_column = "DATE(submission_timestamp)"
            client_id_column = "client_info.client_id"


            [dimensions]

            [dimensions.os]
            data_source = "events"
            select_expression = "normalized_os"
            friendly_name = "Operating System"
            description = "Normalized Operating System"
            client_id_column = "client_info.client_id"
            """
        )
        (config_dir / "test-funnel.toml").write_text(config)

        generate_funnels(
            target_project="test-project",
            path=config_dir,
            output_dir=output_dir.parent.parent,
        )

        (output_dir / "test_funnel_v1" / "query.sql").read_text() == (
            BASE_DIR / "tests" / "test_funnel"
        ).read_text()



import distutils
import os
import tempfile
from datetime import datetime
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml
from click.testing import CliRunner
from dateutil.relativedelta import relativedelta

from bigquery_etl.cli.metadata import deprecate, publish, update
from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.metadata.validate_metadata import (
    validate,
    validate_change_control,
    validate_shredder_mitigation,
)

TEST_DIR = Path(__file__).parent.parent


class TestMetadata:
    test_path = "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1"

    @property
    def codeowners_filename(self) -> str:
        return "CODEOWNERS_TEST"

    @pytest.fixture
    def runner(self):
        return CliRunner()

    def check_metadata(
        self, runner, metadata_conf, codeowners_conf=None, expected_result=None
    ):
        codeowners_file = os.path.join(self.test_path, self.codeowners_filename)
        with runner.isolated_filesystem():
            os.makedirs(self.test_path)
            with open(
                f"{self.test_path}/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            metadata = Metadata.from_file(f"{self.test_path}/metadata.yaml")
            with open(codeowners_file, "w") as codeowners:
                codeowners.write(codeowners_conf or "")

            assert self.codeowners_filename in os.listdir(self.test_path)
            assert (
                validate_change_control(
                    file_path=self.test_path,
                    metadata=metadata,
                    codeowners_file=codeowners_file,
                )
                is expected_result
            )

    def test_validate_change_control_no_owners_in_codeowners(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": ["test@example.org"],
            "labels": {"change_controlled": "true", "foo": "abc"},
        }
        self.check_metadata(
            runner=runner, metadata_conf=metadata, expected_result=False
        )

    def test_validate_change_control_no_owners_in_metadata(self, runner):
        metadata = {
            "friendly_name": "test",
            "labels": {"change_controlled": "true", "foo": "abc"},
        }
        self.check_metadata(
            runner=runner, metadata_conf=metadata, expected_result=False
        )

    def test_validate_change_control_ok(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": ["test@example.org"],
            "labels": {"change_controlled": "true"},
        }
        codeowners = (
            "/sql/moz-fx-data-shared-prod/telemetry_derived/query_v1 test@example.org"
        )
        self.check_metadata(
            runner=runner,
            metadata_conf=metadata,
            codeowners_conf=codeowners,
            expected_result=True,
        )

    def test_validate_change_control_at_lest_one_owner(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": [
                "test@example.org",
                "test2@example.org",
                "test3@example.org",
            ],
            "labels": {"change_controlled": "true", "foo": "abc"},
        }
        codeowners = (
            "/sql/moz-fx-data-shared-prod/telemetry_derived/query_v1 test2@example.org"
        )
        self.check_metadata(
            runner=runner,
            metadata_conf=metadata,
            codeowners_conf=codeowners,
            expected_result=True,
        )

    def test_validate_change_control_all_owners(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": [
                "test@example.org",
                "test2@example.org",
                "test3@example.org",
            ],
            "labels": {"change_controlled": "true", "foo": "abc"},
        }
        codeowners = "/sql/moz-fx-data-shared-prod/telemetry_derived/query_v1 test@example.org test2@example.org test3@example.org"
        self.check_metadata(
            runner=runner,
            metadata_conf=metadata,
            codeowners_conf=codeowners,
            expected_result=True,
        )

    def test_validate_change_control_github_identity(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": [
                "test@example.org",
                "test2@example.org",
                "mozilla/reviewers",
            ],
            "labels": {"change_controlled": "true", "foo": "abc"},
        }
        codeowners = (
            "/sql/moz-fx-data-shared-prod/telemetry_derived/query_v1 @mozilla/reviewers"
        )
        self.check_metadata(
            runner=runner,
            metadata_conf=metadata,
            codeowners_conf=codeowners,
            expected_result=True,
        )

    def test_validate_change_control_no_label(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": ["test@example.org"],
            "labels": {"foo": "abc"},
        }
        self.check_metadata(runner=runner, metadata_conf=metadata, expected_result=True)

    def test_validate_change_control_more_than_one_owner_but_not_all(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": ["test@example.org", "test2@example.org", "test3@example.org"],
            "labels": {"change_controlled": "true"},
        }
        codeowners = "/sql/moz-fx-data-shared-prod/telemetry_derived/query_v1 test@example.org test2@example.org"
        self.check_metadata(
            runner=runner,
            metadata_conf=metadata,
            codeowners_conf=codeowners,
            expected_result=True,
        )

    def test_validate_change_control_commented_line(self, runner):
        metadata = {
            "friendly_name": "test",
            "owners": ["test@example.org", "test2@example.org"],
            "labels": {"change_controlled": "true"},
        }
        codeowners = (
            "#/sql/moz-fx-data-shared-prod/telemetry_derived/query_v1 test@example.org"
        )
        self.check_metadata(
            runner=runner,
            metadata_conf=metadata,
            codeowners_conf=codeowners,
            expected_result=False,
        )

    def test_metadata_update_with_no_deprecation(self, runner):
        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))
            name = [
                str(tmpdirname)
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_v6/"
            ]
            runner.invoke(update, name, "--sql_dir=" + str(tmpdirname) + "/sql")
            with open(
                tmpdirname
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_v6/metadata.yaml",
                "r",
            ) as stream:
                metadata = yaml.safe_load(stream)
        assert metadata["workgroup_access"][0]["role"] == "roles/bigquery.dataViewer"
        assert metadata["workgroup_access"][0]["members"] == [
            "workgroup:mozilla-confidential"
        ]
        assert "deprecated" not in metadata

    def test_metadata_update_with_deprecation(self, runner):
        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))
            name = [
                str(tmpdirname)
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_scalar_aggregates_v1/"
            ]
            runner.invoke(update, name, "--sql_dir=" + str(tmpdirname) + "/sql")
            with open(
                tmpdirname
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_scalar_aggregates_v1/metadata.yaml",
                "r",
            ) as stream:
                metadata = yaml.safe_load(stream)

            with open(
                tmpdirname
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/dataset_metadata.yaml",
                "r",
            ) as stream:
                dataset_metadata = yaml.safe_load(stream)

        assert metadata["workgroup_access"] == []
        assert metadata["deprecated"]
        assert dataset_metadata["workgroup_access"] == [
            {
                "members": ["workgroup:mozilla-test"],
                "role": "roles/bigquery.dataEditor",
            }
        ]
        assert dataset_metadata["default_table_workgroup_access"] == [
            {
                "members": ["workgroup:mozilla-confidential"],
                "role": "roles/bigquery.dataViewer",
            }
        ]

    def test_metadata_update_do_not_update(self, runner):
        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))
            name = [
                str(tmpdirname)
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_scalar_aggregates_v1/"
            ]
            runner.invoke(update, name, "--sql_dir=" + str(tmpdirname) + "/sql")
            with open(
                tmpdirname
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_scalar_aggregates_v1/metadata.yaml",
                "r",
            ) as stream:
                metadata = yaml.safe_load(stream)

        assert metadata["workgroup_access"][0]["role"] == "roles/bigquery.dataViewer"
        assert metadata["workgroup_access"][0]["members"] == ["workgroup:revenue/cat4"]
        assert "deprecated" not in metadata

    @patch("google.cloud.bigquery.Client")
    @patch("google.cloud.bigquery.Table")
    def test_metadata_publish(self, mock_bigquery_table, mock_bigquery_client, runner):
        mock_bigquery_client().get_table.return_value = mock_bigquery_table()
        from distutils import log

        log.set_verbosity(log.WARN)
        log.set_threshold(log.WARN)

        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))
            name = (
                str(tmpdirname)
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_scalar_aggregates_v1/"
            )

            runner.invoke(
                publish,
                [name, "--parallelism=0", "--sql_dir=" + str(tmpdirname) + "/sql"],
                catch_exceptions=False,
            )

        assert mock_bigquery_client().update_table.call_count == 1
        assert (
            mock_bigquery_client().update_table.call_args[0][0].friendly_name
            == "Test metadata.yaml"
        )
        assert (
            mock_bigquery_client().update_table.call_args[0][0].description
            == "Clustering fields: `column1`"
        )
        assert mock_bigquery_client().update_table.call_args[0][0].labels == {
            "deletion_date": "2024-03-02",
            "deprecated": "true",
            "owner1": "test",
            "monitoring": "true",
        }
        assert (
            mock_bigquery_client()
            .get_table(
                "moz-fx-data-shared-prod.telemetry_derived.clients_daily_scalar_aggregates_v1"
            )
            .friendly_name
            == "Test metadata.yaml"
        )
        assert mock_bigquery_table().friendly_name == "Test metadata.yaml"
        assert mock_bigquery_table().description == "Clustering fields: `column1`"
        assert mock_bigquery_table().labels == {
            "deletion_date": "2024-03-02",
            "deprecated": "true",
            "owner1": "test",
            "monitoring": "true",
        }

    @patch("google.cloud.bigquery.Client")
    @patch("google.cloud.bigquery.Table")
    def test_metadata_publish_with_no_metadata_file(
        self, mock_bigquery_table, mock_bigquery_client, runner
    ):
        mock_bigquery_client().get_table.return_value = mock_bigquery_table()

        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))
            name = [
                str(tmpdirname)
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_scalar_aggregates_v2/"
            ]
            runner.invoke(publish, name, "--sql_dir=" + str(tmpdirname) + "/sql")

        assert mock_bigquery_client().update_table.call_count == 0

    def test_metadata_deprecate_default_deletion_date(self, runner):
        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))

            qualified_table_name = (
                "moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6"
            )
            result = runner.invoke(
                deprecate,
                [qualified_table_name, "--sql_dir=" + str(tmpdirname) + "/sql"],
            )
            with open(
                tmpdirname
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_v6/metadata.yaml",
                "r",
            ) as stream:
                metadata = yaml.safe_load(stream)

        default_deletion_date = (datetime.today() + relativedelta(months=+3)).date()

        assert result.exit_code == 0
        assert metadata["deprecated"]
        assert metadata["deletion_date"] == default_deletion_date

    def test_metadata_deprecate_set_deletion_date(self, runner):
        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))

            qualified_table_name = (
                "moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6"
            )
            result = runner.invoke(
                deprecate,
                [
                    qualified_table_name,
                    "--deletion_date=2024-03-02",
                    "--sql_dir=" + str(tmpdirname) + "/sql",
                ],
            )
            with open(
                tmpdirname
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_v6/metadata.yaml",
                "r",
            ) as stream:
                metadata = yaml.safe_load(stream)

        assert result.exit_code == 0
        assert metadata["deprecated"]
        assert metadata["deletion_date"] == datetime(2024, 3, 2).date()

    def test_metadata_deprecate_set_invalid_deletion_date_should_fail(self, runner):
        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))

            qualified_table_name = (
                "moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6"
            )
            result = runner.invoke(
                deprecate,
                [
                    qualified_table_name,
                    "--deletion_date=2024-02",
                    "--sql_dir=" + str(tmpdirname) + "/sql",
                ],
            )
            with open(
                tmpdirname
                + "/sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_v6/metadata.yaml",
                "r",
            ) as stream:
                metadata = yaml.safe_load(stream)

        assert result.exit_code == 2
        assert "deprecated" not in metadata
        assert "deletion_date" not in metadata
        assert "Invalid value for '--deletion_date'" in result.output

    def test_metadata_deprecate_no_metadata(self, runner):
        with tempfile.TemporaryDirectory() as tmpdirname:
            distutils.dir_util.copy_tree(str(TEST_DIR), str(tmpdirname))

            qualified_table_name = "moz-fx-data-shared-prod.telemetry_derived.clients_daily_scalar_aggregates_v2"
            result = runner.invoke(
                deprecate,
                [
                    qualified_table_name,
                    "--deletion_date=2024-03-02",
                    "--sql_dir=" + str(tmpdirname) + "/sql",
                ],
            )

            assert result.exit_code == 1
            assert (
                str(result.exception)
                == f"No metadata file(s) were found for: {qualified_table_name}"
            )

    def test_validate_shredder_mitigation_schema_columns(self, runner, capfd):
        """Test that validation fails when query contains id-level columns or descriptions are missing."""
        metadata = {
            "friendly_name": "Test",
            "labels": {"shredder_mitigation": "true"},
        }
        schema_1 = {
            "fields": [
                {
                    "mode": "REQUIRED",
                    "name": "profile_id",
                    "type": "STRING",
                    "description": "description 1",
                },
                {
                    "mode": "REQUIRED",
                    "name": "column_3",
                    "type": "STRING",
                    "description": "description 3",
                },
            ]
        }
        schema_2 = {
            "fields": [
                {
                    "mode": "NULLABLE",
                    "name": "column_1",
                    "type": "STRING",
                },
                {
                    "name": "column_2",
                    "type": "STRING",
                    "description": "description 3",
                },
            ]
        }

        with runner.isolated_filesystem():
            query_path = Path(self.test_path) / "query.sql"
            metadata_path = Path(self.test_path) / "metadata.yaml"
            schema_path = Path(self.test_path) / "schema.yaml"
            os.makedirs(self.test_path, exist_ok=True)

            with open(query_path, "w") as f:
                f.write("SELECT column_1 FROM test_table group by column_1")
            with open(metadata_path, "w") as f:
                f.write(yaml.safe_dump(metadata))
            with open(schema_path, "w") as f:
                f.write(yaml.safe_dump(schema_1))
            metadata_from_file = Metadata.from_file(metadata_path)

            result = validate_shredder_mitigation(self.test_path, metadata_from_file)
            captured = capfd.readouterr()
            assert result is False
            assert (
                "Shredder mitigation validation failed, profile_id is an id-level column "
                "that is not allowed for this type of backfill."
            ) in captured.out

            with open(schema_path, "w") as f:
                f.write(yaml.safe_dump(schema_2))
            result = validate_shredder_mitigation(self.test_path, metadata_from_file)
            captured = capfd.readouterr()
            assert result is False
            assert (
                "Shredder mitigation validation failed, column_1 does not have a description"
                " in the schema."
            ) in captured.out

    def test_validate_shredder_mitigation_group_by(self, runner, capfd):
        """Test that validation fails and prints a notification when group by contains numbers."""
        metadata = {
            "friendly_name": "Test",
            "labels": {"shredder_mitigation": "true"},
        }

        schema = {
            "fields": [
                {
                    "mode": "REQUIRED",
                    "name": "column_1",
                    "type": "STRING",
                },
                {
                    "mode": "REQUIRED",
                    "name": "column_2",
                    "type": "STRING",
                    "description": "description 2",
                },
            ]
        }

        with runner.isolated_filesystem():
            query_path = Path(self.test_path) / "query.sql"
            metadata_path = Path(self.test_path) / "metadata.yaml"
            schema_path = Path(self.test_path) / "schema.yaml"
            os.makedirs(self.test_path, exist_ok=True)

            with open(query_path, "w") as f:
                f.write("SELECT column_1 FROM test_table GROUP BY ALL")
            with open(metadata_path, "w") as f:
                f.write(yaml.safe_dump(metadata))
            with open(schema_path, "w") as f:
                f.write(yaml.safe_dump(schema))

            metadata_from_file = Metadata.from_file(metadata_path)
            result = validate_shredder_mitigation(self.test_path, metadata_from_file)
            captured = capfd.readouterr()
            assert result is False
            assert (
                "Shredder mitigation validation failed, GROUP BY must use an explicit list of"
                " columns. Avoid expressions like `GROUP BY ALL` or `GROUP BY 1, 2, 3`."
            ) in captured.out

            with open(query_path, "w") as f:
                f.write("SELECT column_1 FROM test_table GROUP BY column_1, 2")
            result = validate_shredder_mitigation(self.test_path, metadata_from_file)
            captured = capfd.readouterr()
            assert result is False
            assert (
                "Shredder mitigation validation failed, GROUP BY must use an explicit list "
                "of columns. Avoid expressions like `GROUP BY ALL` or `GROUP BY 1, 2, 3`."
            ) in captured.out

    def test_validate_metadata_without_labels(self, runner, capfd):
        """Test that metadata validation doesn't fail when labels are not present."""
        metadata = {
            "friendly_name": "Test",
            "labels": {},
        }
        schema = """
            fields:
              - mode: NULLABLE
                name: column_1
                type: DATE
            """

        with runner.isolated_filesystem():
            os.makedirs(self.test_path, exist_ok=True)
            with open(Path(self.test_path) / "metadata.yaml", "w") as f:
                f.write(yaml.safe_dump(metadata))
            with open(Path(self.test_path) / "schema.yaml", "w") as f:
                f.write(schema)

            result = validate(self.test_path)
            captured = capfd.readouterr()
            assert result is None
            assert captured.out == ""



"""Threaded TopologialSorter."""

import multiprocessing
from copy import deepcopy
from graphlib import TopologicalSorter
from typing import Dict, Set


class ParallelTopologicalSorter:
    """TopologicalSorter that processes sorted items concurrently."""

    def __init__(
        self,
        dependencies: Dict[str, Set[str]],
        parallelism: int = 8,
        with_follow_up=False,
    ):
        """Initialize."""
        self.dependencies = deepcopy(dependencies)
        self.parallelism = parallelism

        manager = multiprocessing.Manager()
        self.visited = manager.dict({dep: False for dep, _ in dependencies.items()})
        self.with_follow_up = with_follow_up

    def _worker(
        self, task_queue, finalized_tasks_queue, followup_queue, visited_map, callback
    ):
        while True:
            item = task_queue.get()
            callback(item, followup_queue)
            if item in visited_map:
                visited_map[item] = True
            finalized_tasks_queue.put(item)
            task_queue.task_done()

    def map(self, callback):
        """Sort and process dependencies."""
        task_queue = multiprocessing.JoinableQueue()
        finalized_tasks_queue = multiprocessing.JoinableQueue()
        processes = []

        if self.with_follow_up:
            followup_queue = multiprocessing.Queue()
        else:
            followup_queue = None

        for _ in range(self.parallelism):
            process = multiprocessing.Process(
                target=self._worker,
                args=(
                    task_queue,
                    finalized_tasks_queue,
                    followup_queue,
                    self.visited,
                    callback,
                ),
                daemon=True,
            )
            process.start()
            processes.append(process)

        # pallel sorting and processing
        ts = TopologicalSorter(self.dependencies)
        ts.prepare()

        while ts.is_active():
            # queue task to process
            for task in ts.get_ready():
                task_queue.put(task)

            # task done
            task = finalized_tasks_queue.get()
            ts.done(task)
            finalized_tasks_queue.task_done()

        task_queue.join()
        finalized_tasks_queue.join()

        # followup items cannot be added to the topological sorter after prepare() is called
        # processing them here sequentially
        if followup_queue:
            while not followup_queue.empty():
                follow_up_item = followup_queue.get()
                callback(follow_up_item, followup_queue)

        # check that all dependencies have been processed
        for task, _ in self.visited.items():
            assert self.visited[task]



"""Utility functions related to UDF parsing and documentation."""

import re
from typing import Optional, Tuple


def clean_sql_text(sql_text: str) -> str:
    """Remove line breaks, comments and multiple whitespaces from sql_text."""
    sql_text_without_comments = re.sub(r"--.+$", "", sql_text, flags=re.MULTILINE)
    return re.sub(r"\s{2,}|\n", " ", sql_text_without_comments).strip()


def clean_result(result_string: str) -> str:
    """Remove white spaces around brackets."""
    return re.sub(r"(?<=\<|\>)\s+|\s+(?=\<|\>)", "", result_string)


def get_input(sql_text: str) -> Optional[str]:
    """Extract input parameters from UDF sql code."""
    cleaned_sql_text = clean_sql_text(sql_text)
    input_groups = re.search(
        r"(?:FUNCTION|PROCEDURE) [a-z,0-9,_,.]+\s?\(\s?(.+?)(?:\s?\)|, OUT)",
        cleaned_sql_text,
    )
    if input_groups is None:
        return None
    input_string = input_groups.groups()[0]
    return clean_result(input_string)


def get_output(sql_text: str) -> Optional[str]:
    """Extract output parameters from UDF sql code."""
    cleaned_sql_text = clean_sql_text(sql_text)
    output_groups = re.search(
        r"(?:RETURNS |CAST\([a-z,A-Z,0-9,_,.,\-,\s,&,\(\),>]+ AS |, OUT )(.+?)(?: AS|\s?\)|\s?LANGUAGE)",
        cleaned_sql_text,
    )
    if output_groups is None:
        return None
    output_string = output_groups.groups()[0]
    return clean_result(output_string)


def get_mozfun_parameters(sql_text: str) -> Tuple:
    """Extract input and output parameters from UDF sql code and stored procedure sql code.

    The sql_text parameter can include sql codes header, body, tests & comments.
    """
    input_part = get_input(sql_text)
    output_part = get_output(sql_text)
    return input_part, output_part



"""Import Apple reports into BigQuery."""

import gzip
import sys
import time
import warnings
from datetime import datetime
from io import BytesIO
from tempfile import TemporaryFile
from typing import Optional

import requests
import rich_click as click
from authlib.jose import jwt
from google.cloud import bigquery

SUBSCRIBER_SCHEMA = [
    bigquery.SchemaField("report_date", "DATE"),
    bigquery.SchemaField("event_date", "DATE"),
    bigquery.SchemaField("app_name", "STRING"),
    bigquery.SchemaField("app_apple_id", "INT64"),
    bigquery.SchemaField("subscription_name", "STRING"),
    bigquery.SchemaField("subscription_apple_id", "INT64"),
    bigquery.SchemaField("subscription_group_id", "INT64"),
    bigquery.SchemaField("standard_subscription_duration", "STRING"),
    bigquery.SchemaField("subscription_offer_name", "STRING"),
    bigquery.SchemaField("promotional_offer_id", "STRING"),
    bigquery.SchemaField("subscription_offer_type", "STRING"),
    bigquery.SchemaField("subscription_offer_duration", "STRING"),
    bigquery.SchemaField("marketing_opt_in_duration", "STRING"),
    bigquery.SchemaField("customer_price", "NUMERIC"),
    bigquery.SchemaField("customer_currency", "STRING"),
    bigquery.SchemaField("developer_proceeds", "NUMERIC"),
    bigquery.SchemaField("proceeds_currency", "STRING"),
    bigquery.SchemaField("preserved_pricing", "STRING"),
    bigquery.SchemaField("proceeds_reason", "STRING"),
    bigquery.SchemaField("client", "STRING"),
    bigquery.SchemaField("device", "STRING"),
    bigquery.SchemaField("country", "STRING"),
    bigquery.SchemaField("subscriber_id", "INT64"),
    bigquery.SchemaField("subscriber_id_reset", "STRING"),
    bigquery.SchemaField("refund", "STRING"),
    bigquery.SchemaField("purchase_date", "DATE"),
    bigquery.SchemaField("units", "NUMERIC"),
]


def _get_lines(
    key_id: Optional[str],
    issuer_id: Optional[str],
    private_key: Optional[str],
    vendor_number: Optional[str],
    report_date: Optional[datetime],
):
    if not all((key_id, issuer_id, private_key, vendor_number)):
        yield from sys.stdin.buffer
    else:
        expiration_time = int(round(time.time() + (20 * 60)))  # 20 minutes
        token = jwt.encode(
            {"alg": "ES256", "kid": key_id, "typ": "JWT"},
            {"iss": issuer_id, "exp": expiration_time, "aud": "appstoreconnect-v1"},
            private_key,
        )
        try:
            response = requests.get(
                "https://api.appstoreconnect.apple.com/v1/salesReports",
                params={
                    "filter[frequency]": "DAILY",
                    "filter[reportDate]": f"{report_date:%F}",
                    "filter[reportSubType]": "DETAILED",
                    "filter[reportType]": "SUBSCRIBER",
                    "filter[vendorNumber]": vendor_number,
                    "filter[version]": "1_3",
                },
                headers={"Authorization": "Bearer " + token.decode()},
            )
        except requests.RequestException as e:
            raise click.ClickException(str(e))
        try:
            response.raise_for_status()
        except requests.RequestException as e:
            raise click.ClickException(f"{e}\n" + response.content.decode("UTF-8"))
        yield from BytesIO(gzip.decompress(response.content))


@click.group(help="Commands for Apple Reports ETL.")
def apple():
    """CLI group for apple commands."""
    pass


@apple.command("import")
@click.option(
    "--key-id",
    help="API JWT key id; Must be specified with --issuer-id, --private-key, and "
    "--vendor-number; If not set rows will be read from stdin",
)
@click.option(
    "--issuer-id",
    help="API JWT issuer id; Must be specified with --key-id, --private-key, and "
    "--vendor-number; If not set rows will be read from stdin",
)
@click.option(
    "--private-key",
    help="API JWT private key; Must be specified with --key-id, --issuer-id, and "
    "--vendor-number; If not set rows will be read from stdin",
)
@click.option(
    "--vendor-number",
    help="Apple vendor number; Must be specified with --key-id, --issuer-id, and "
    "--private-key; If not set rows will be read from stdin",
)
@click.option(
    "--date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Report date to pull from API; Added to --table to ensure only that date "
    "partition is replaced",
)
@click.option(
    "--table",
    help="BigQuery Standard SQL format table ID where resources will be written; if "
    "not set resources will be written to stdout",
)
def apple_import(
    key_id: Optional[str],
    issuer_id: Optional[str],
    private_key: Optional[str],
    vendor_number: Optional[str],
    date: Optional[datetime],
    table: Optional[str],
):
    """Import Apple Subscriber report into BigQuery."""
    auth_options = (key_id, issuer_id, private_key, vendor_number)
    if any(auth_options) and not all(auth_options):
        raise click.ClickException(
            "Must set all of --key-id, --issuer-id, --private-key, and "
            "--vendor-number or none of them"
        )
    if (table or key_id) and not date:
        raise click.ClickException("must specify --date")
    if table and date:
        table = f"{table}${date:%Y%m%d}"
    with TemporaryFile(mode="w+b") if table else sys.stdout.buffer as file_obj:
        has_rows = False
        report_date = f"{date:%F}\t".encode("UTF-8")
        for line in _get_lines(key_id, issuer_id, private_key, vendor_number, date):
            file_obj.write(report_date)  # prefix lines with report date column
            file_obj.write(line)
            has_rows = True
        if not has_rows:
            raise click.ClickException("no rows returned")
        elif table:
            if file_obj.writable():
                file_obj.seek(0)
            warnings.filterwarnings("ignore", module="google.auth._default")
            job = bigquery.Client().load_table_from_file(
                file_obj=file_obj,
                destination=table,
                job_config=bigquery.LoadJobConfig(
                    allow_jagged_rows=False,
                    field_delimiter="\t",
                    ignore_unknown_values=False,
                    schema=SUBSCRIBER_SCHEMA,  # only report type supported
                    schema_update_options=[
                        bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION
                    ],
                    skip_leading_rows=1,
                    source_format=bigquery.SourceFormat.CSV,
                    time_partitioning=bigquery.TimePartitioning(field="report_date"),
                    write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
                ),
            )
            try:
                click.echo(f"Waiting for {job.job_id}", file=sys.stderr)
                job.result()
            except Exception as e:
                message = f"{job.job_id} failed: {e}"
                for error in job.errors or ():
                    error_message = error.get("message")
                    if error_message and error_message != getattr(e, "message", None):
                        message += f"\n{error_message}"
                raise click.ClickException(message)
            else:
                click.echo(f"{job.job_id} succeeded", file=sys.stderr)



#!/usr/bin/env python3

"""Import experiments from Experimenter via the Experimenter API."""

import datetime
import json
import sys
import time
from argparse import ArgumentParser
from typing import List, Optional

import attr
import cattrs
import pytz
import requests
from google.cloud import bigquery

EXPERIMENTER_API_URL_V1 = (
    "https://experimenter.services.mozilla.com/api/v1/experiments/"
)

# for nimbus experiments
EXPERIMENTER_API_URL_V6 = (
    "https://experimenter.services.mozilla.com/api/v6/experiments/"
)

parser = ArgumentParser(description=__doc__)
parser.add_argument("--project", default="moz-fx-data-experiments")
parser.add_argument("--destination_dataset", default="monitoring")
parser.add_argument("--destination_table", default="experimenter_experiments_v1")
parser.add_argument("--sql_dir", default="sql/")
parser.add_argument("--dry_run", action="store_true")


@attr.s(auto_attribs=True)
class Branch:
    """Defines a branch."""

    slug: str
    ratio: int
    features: Optional[dict]


@attr.s(auto_attribs=True)
class Experiment:
    """Defines an Experiment."""

    experimenter_slug: Optional[str]
    normandy_slug: Optional[str]
    type: str
    status: Optional[str]
    branches: List[Branch]
    start_date: Optional[datetime.datetime]
    end_date: Optional[datetime.datetime]
    enrollment_end_date: Optional[datetime.datetime]
    proposed_enrollment: Optional[int]
    reference_branch: Optional[str]
    is_high_population: bool
    app_name: str
    app_id: str
    channel: str
    targeting: str
    targeted_percent: float
    namespace: Optional[str]
    feature_ids: List[str]


def _coerce_none_to_zero(x: Optional[int]) -> int:
    return 0 if x is None else x


@attr.s(auto_attribs=True)
class Variant:
    """Defines a Variant."""

    is_control: bool
    slug: str
    ratio: int
    value: str


@attr.s(auto_attribs=True)
class ExperimentV1:
    """Experimenter v1 experiment."""

    slug: str  # experimenter slug
    type: str
    status: str
    start_date: Optional[datetime.datetime]
    end_date: Optional[datetime.datetime]
    variants: List[Variant]
    proposed_enrollment: Optional[int] = attr.ib(converter=_coerce_none_to_zero)
    firefox_channel: str
    population_percent: float
    normandy_slug: Optional[str] = None
    is_high_population: Optional[bool] = None

    @staticmethod
    def _unix_millis_to_datetime(num: Optional[float]) -> Optional[datetime.datetime]:
        """Convert unix milliseconds to datetime."""
        if num is None:
            return None
        return datetime.datetime.fromtimestamp(num / 1e3, pytz.utc)

    @classmethod
    def from_dict(cls, d) -> "ExperimentV1":
        """Load an experiment from dict."""
        converter = cattrs.BaseConverter()
        converter.register_structure_hook(
            datetime.datetime,
            lambda num, _: cls._unix_millis_to_datetime(num),
        )
        return converter.structure(d, cls)

    def to_experiment(self) -> "Experiment":
        """Convert to Experiment."""
        branches = [
            Branch(slug=variant.slug, ratio=variant.ratio, features=None)
            for variant in self.variants
        ]
        control_slug = None

        control_slugs = [
            variant.slug for variant in self.variants if variant.is_control
        ]
        if len(control_slugs) == 1:
            control_slug = control_slugs[0]

        return Experiment(
            normandy_slug=self.normandy_slug,
            experimenter_slug=self.slug,
            type=self.type,
            status=self.status,
            start_date=self.start_date,
            end_date=self.end_date,
            enrollment_end_date=None,
            branches=branches,
            proposed_enrollment=self.proposed_enrollment,
            reference_branch=control_slug,
            is_high_population=self.is_high_population or False,
            app_name="firefox_desktop",
            app_id="firefox-desktop",
            channel=self.firefox_channel.lower(),
            targeting="",
            targeted_percent=float(self.population_percent) / 100.0,
            namespace=None,
            feature_ids=[],
        )


@attr.s(auto_attribs=True)
class ExperimentV6:
    """Represents a v6 experiment from Experimenter."""

    slug: str  # Normandy slug
    startDate: Optional[datetime.datetime]
    endDate: Optional[datetime.datetime]
    enrollmentEndDate: Optional[datetime.datetime]
    proposedEnrollment: int
    branches: List[Branch]
    referenceBranch: Optional[str]
    appName: str
    appId: str
    channel: str
    targeting: str
    bucketConfig: dict
    featureIds: list[str]

    @classmethod
    def from_dict(cls, d) -> "ExperimentV6":
        """Load an experiment from dict."""
        converter = cattrs.BaseConverter()
        converter.register_structure_hook(
            datetime.datetime,
            lambda num, _: datetime.datetime.fromisoformat(
                num.replace("Z", "+00:00")
            ).astimezone(pytz.utc),
        )
        converter.register_structure_hook(
            Branch,
            lambda b, _: Branch(
                slug=b["slug"], ratio=b["ratio"], features=b["features"]
            ),
        )
        return converter.structure(d, cls)

    def to_experiment(self) -> "Experiment":
        """Convert to Experiment."""
        return Experiment(
            normandy_slug=self.slug,
            experimenter_slug=None,
            type="v6",
            status=(
                "Live"
                if (
                    self.endDate is None
                    or (
                        self.endDate
                        and self.endDate > pytz.utc.localize(datetime.datetime.now())
                    )
                )
                else "Complete"
            ),
            start_date=self.startDate,
            end_date=self.endDate,
            enrollment_end_date=self.enrollmentEndDate,
            proposed_enrollment=self.proposedEnrollment,
            reference_branch=self.referenceBranch,
            is_high_population=False,
            branches=self.branches,
            app_name=self.appName,
            app_id=self.appId,
            channel=self.channel,
            targeting=self.targeting,
            targeted_percent=self.bucketConfig["count"] / self.bucketConfig["total"],
            namespace=self.bucketConfig["namespace"],
            feature_ids=self.featureIds,
        )


def fetch(url):
    """Fetch a url."""
    for _ in range(2):
        try:
            return requests.get(
                url,
                timeout=30,
                headers={"user-agent": "https://github.com/mozilla/bigquery-etl"},
            ).json()
        except Exception as e:
            last_exception = e
            time.sleep(1)
    raise last_exception


def get_experiments() -> List[Experiment]:
    """Fetch experiments from Experimenter."""
    legacy_experiments_json = fetch(EXPERIMENTER_API_URL_V1)
    legacy_experiments = []

    for experiment in legacy_experiments_json:
        if experiment["type"] != "rapid":
            try:
                legacy_experiments.append(
                    ExperimentV1.from_dict(experiment).to_experiment()
                )
            except Exception as e:
                print(f"Cannot import experiment: {experiment}: {e}")

    nimbus_experiments_json = fetch(EXPERIMENTER_API_URL_V6)
    nimbus_experiments = []

    for experiment in nimbus_experiments_json:
        try:
            nimbus_experiments.append(
                ExperimentV6.from_dict(experiment).to_experiment()
            )
        except Exception as e:
            print(f"Cannot import experiment: {experiment}: {e}")

    return nimbus_experiments + legacy_experiments


def main():
    """Run."""
    args = parser.parse_args()
    experiments = get_experiments()

    destination_table = (
        f"{args.project}.{args.destination_dataset}.{args.destination_table}"
    )

    bq_schema = (
        bigquery.SchemaField("experimenter_slug", "STRING"),
        bigquery.SchemaField("normandy_slug", "STRING"),
        bigquery.SchemaField("type", "STRING"),
        bigquery.SchemaField("status", "STRING"),
        bigquery.SchemaField("start_date", "DATE"),
        bigquery.SchemaField("end_date", "DATE"),
        bigquery.SchemaField("enrollment_end_date", "DATE"),
        bigquery.SchemaField("proposed_enrollment", "INTEGER"),
        bigquery.SchemaField("reference_branch", "STRING"),
        bigquery.SchemaField("is_high_population", "BOOL"),
        bigquery.SchemaField(
            "branches",
            "RECORD",
            mode="REPEATED",
            fields=[
                bigquery.SchemaField("slug", "STRING"),
                bigquery.SchemaField("ratio", "INTEGER"),
                bigquery.SchemaField("features", "JSON"),
            ],
        ),
        bigquery.SchemaField("app_id", "STRING"),
        bigquery.SchemaField("app_name", "STRING"),
        bigquery.SchemaField("channel", "STRING"),
        bigquery.SchemaField("targeting", "STRING"),
        bigquery.SchemaField("targeted_percent", "FLOAT"),
        bigquery.SchemaField("namespace", "STRING"),
        bigquery.SchemaField("feature_ids", "STRING", mode="REPEATED"),
    )

    job_config = bigquery.LoadJobConfig(
        write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
    )
    job_config.schema = bq_schema

    converter = cattrs.BaseConverter()
    converter.register_unstructure_hook(
        datetime.datetime, lambda d: datetime.datetime.strftime(d, format="%Y-%m-%d")
    )

    blob = converter.unstructure(experiments)
    if args.dry_run:
        print(json.dumps(blob))
        sys.exit(0)

    client = bigquery.Client(args.project)
    client.load_table_from_json(blob, destination_table, job_config=job_config).result()
    print(f"Loaded {len(experiments)} experiments")


if __name__ == "__main__":
    main()



"""Generate Aggregate table for monitoring event flows."""

import os
from collections import namedtuple
from pathlib import Path

from sql_generators.glean_usage.common import (
    GleanTable,
    get_table_dir,
    render,
    write_sql,
)

AGGREGATE_TABLE_NAME = "event_flow_monitoring_aggregates_v1"
TARGET_DATASET_CROSS_APP = "monitoring"
PREFIX = "event_flow_monitoring"
PATH = Path(os.path.dirname(__file__))


class EventFlowMonitoring(GleanTable):
    """Represents the generated aggregated table for event flow monitoring."""

    def __init__(self) -> None:
        self.per_app_id_enabled = False
        self.per_app_enabled = False
        self.across_apps_enabled = True
        self.prefix = PREFIX
        self.target_table_id = AGGREGATE_TABLE_NAME
        self.custom_render_kwargs = {}
        self.base_table_name = "events_unnested"

    def generate_across_apps(
        self, project_id, apps, output_dir=None, use_cloud_function=True, parallelism=8
    ):
        """Generate a query across all apps."""
        if not self.across_apps_enabled:
            return

        apps = [app[0] for app in apps]

        render_kwargs = dict(
            project_id=project_id,
            target_table=f"{TARGET_DATASET_CROSS_APP}_derived.{AGGREGATE_TABLE_NAME}",
            apps=apps,
        )
        render_kwargs.update(self.custom_render_kwargs)

        skip_existing_artifacts = self.skip_existing(output_dir, project_id)

        Artifact = namedtuple("Artifact", "table_id basename sql")

        query_filename = f"{AGGREGATE_TABLE_NAME}.script.sql"
        script_sql = render(
            query_filename, template_folder=PATH / "templates", **render_kwargs
        )
        metadata = render(
            f"{AGGREGATE_TABLE_NAME}.metadata.yaml",
            template_folder=PATH / "templates",
            format=False,
            **render_kwargs,
        )
        schema = render(
            f"{AGGREGATE_TABLE_NAME}.schema.yaml",
            template_folder=PATH / "templates",
            format=False,
            **render_kwargs,
        )
        table = (
            f"{project_id}.{TARGET_DATASET_CROSS_APP}_derived.{AGGREGATE_TABLE_NAME}"
        )

        if output_dir:
            artifacts = [
                Artifact(table, "metadata.yaml", metadata),
                Artifact(table, "script.sql", script_sql),
                Artifact(table, "schema.yaml", schema),
            ]

            for artifact in artifacts:
                destination = (
                    get_table_dir(output_dir, artifact.table_id) / artifact.basename
                )
                skip_existing = destination in skip_existing_artifacts

                write_sql(
                    output_dir,
                    artifact.table_id,
                    artifact.basename,
                    artifact.sql,
                    skip_existing=skip_existing,
                )



#!/usr/bin/env python3

"""Raise exception when there are potential new targets to add to shredder."""
from textwrap import dedent

import click
from google.cloud import bigquery
from google.cloud.bigquery import TableReference
from typing import Dict, List

NEW_TARGETS_QUERY_TEMPLATE = """
SELECT
    *
FROM
    `{source_table}`
WHERE
    run_date = '{run_date}'
"""


def format_sources(sources: List[Dict[str, str]]) -> List[str]:
    return [f"{source['table']}: {source['field']}" for source in sources]


@click.command
@click.option(
    "--run-date", type=click.DateTime(), help="The date to to check for new targets."
)
@click.option(
    "--source-table",
    type=TableReference.from_string,
    metavar="PROJECT.DATASET.TABLE",
    default="moz-fx-data-shared-prod.monitoring_derived.shredder_targets_new_mismatched_v1",
    help="Table to get new shredder targets from, in the form of PROJECT.DATASET.TABLE. "
    "Defaults to "
    "`moz-fx-data-shared-prod.monitoring_derived.shredder_targets_new_mismatched_v1`",
)
def main(run_date, source_table):
    """Find tables in the given project that could be added to shredder."""
    client = bigquery.Client()

    mismatched_tables = list(
        client.query_and_wait(
            NEW_TARGETS_QUERY_TEMPLATE.format(
                source_table=source_table, run_date=run_date.date()
            )
        )
    )

    table_names = []

    for table in mismatched_tables:
        table_name = f"{table.project_id}.{table.dataset_id}.{table.table_id}"
        table_names.append(table_name)
        print(
            dedent(
                f"""
                {table.project_id}.{table.dataset_id}.{table.table_id}: 
                created on {table.table_creation_date}, owners: {table.owners}
                current deletion sources:          {format_sources(table.current_sources)}
                deletion sources based on lineage: {format_sources(table.detected_sources)}
                """
            )
        )

    if len(table_names) > 0:
        raise RuntimeError(
            dedent(
                f"""
                Found {len(table_names)} new tables that may not be configured correctly in shredder: {table_names}
                Verify with table owners if these tables need to be configured in shredder.
                See logs or {source_table} with run_date="{run_date.date()}" for more details.
                """
            )
        )


if __name__ == "__main__":
    main()



import shutil
from pathlib import Path

import click
from jinja2 import FileSystemLoader, Environment

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.schema.stable_table_schema import get_stable_table_schemas
from bigquery_etl.util.common import write_sql, get_table_dir


@click.command("generate")
@click.option(
    "--target-project",
    "--target_project",
    help="Which project the queries should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@use_cloud_function_option
def generate(target_project, output_dir, use_cloud_function):
    stable_datasets = sorted(
        {schema.bq_dataset_family + "_stable" for schema in get_stable_table_schemas()}
    )

    template_dir = Path(__file__).parent / "templates"
    env = Environment(loader=FileSystemLoader(template_dir))

    output_table = f"{target_project}.monitoring_derived.table_partition_expirations_v1"

    template = env.get_template("table_partition_expirations.sql")
    query = template.render(datasets=stable_datasets, destination_table=output_table)

    table_dir = get_table_dir(Path(output_dir) / target_project, output_table)

    write_sql(
        Path(output_dir) / target_project,
        output_table,
        "query.sql",
        reformat(query),
    )

    shutil.copyfile(
        template_dir / "schema.yaml",
        table_dir / "schema.yaml",
    )

    shutil.copyfile(
        template_dir / "metadata.yaml",
        table_dir / "metadata.yaml",
    )


if __name__ == "__main__":
    generate()



"""PyTest plugin for collecting docstyle tests on python scripts."""

import pydocstyle
from pytest_pydocstyle import File, _patch_sys_argv

from . import is_python_executable


# adapted from
# https://github.com/henry0312/pytest-docstyle/blob/v2.2.0/src/pytest_pydocstyle.py#L37-L46
def pytest_collect_file(parent, path):
    """Collect docstyle tests."""
    config = parent.config
    if config.getoption("pydocstyle") and is_python_executable(path):
        parser = pydocstyle.config.ConfigurationParser()
        args = [str(path.basename), "--match", ".*"]
        with _patch_sys_argv(args):
            parser.parse()
        for filename, _, _ in parser.get_files_to_check():
            return File.from_parent(parent=parent, fspath=path, config_parser=parser)



from unittest.mock import patch

import pytest
import yaml
from google.cloud.exceptions import NotFound

from bigquery_etl import deploy


class TestDeploy:

    @pytest.fixture(scope="session")
    def query_path(self, tmp_path_factory):
        path = (
            tmp_path_factory.mktemp("sql")
            / "moz-fx-data-shared-prod"
            / "test"
            / "test_query_v1"
        )
        path.mkdir(parents=True)
        (path / "query.sql").write_text("SELECT 1")
        (path / "schema.yaml").write_text(
            yaml.dump({"fields": [{"name": "f0_", "type": "INTEGER"}]})
        )
        return path

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.dryrun.DryRun")
    def test_deploy_new_table(self, mock_dryrun, mock_client, query_path):

        mock_dryrun().get_schema.return_value = {
            "fields": [{"name": "f0_", "type": "INTEGER"}]
        }

        client = mock_client.return_value
        client.get_table.side_effect = NotFound("table not found")

        deploy.deploy_table(
            artifact_file=query_path / "query.sql",
        )

        client.create_table.assert_called_once()
        assert client.update_table.call_count == 0

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.dryrun.DryRun")
    def test_deploy_new_table_with_missing_dataset(
        self, mock_dryrun, mock_client, query_path
    ):

        mock_dryrun().get_schema.return_value = {
            "fields": [{"name": "f0_", "type": "INTEGER"}]
        }

        client = mock_client.return_value
        client.get_table.side_effect = NotFound("table not found")
        client.create_table.side_effect = NotFound(
            "404 POST https://bigquery.googleapis.com/..."
        )

        with pytest.raises(
            deploy.FailedDeployException, match="Unable to create table"
        ):
            deploy.deploy_table(artifact_file=query_path / "query.sql")

    def test_deploy_table_without_schema_raises_skip(self, tmp_path):
        query_path = (
            tmp_path / "sql" / "moz-fx-data-shared-prod" / "test" / "test_query_v1"
        )
        query_path.mkdir(parents=True)
        (query_path / "query.sql").write_text("SELECT 1")

        with pytest.raises(deploy.SkippedDeployException, match="Schema missing"):
            deploy.deploy_table(artifact_file=query_path / "query.sql")

    def test_deploy_with_null_destination_raises_skip(self, tmp_path):
        query_path = (
            tmp_path / "sql" / "moz-fx-data-shared-prod" / "test" / "test_query_v1"
        )
        query_path.mkdir(parents=True)
        (query_path / "query.sql").write_text("SELECT 1")
        (query_path / "schema.yaml").write_text(
            yaml.dump({"fields": [{"name": "f0_", "type": "INTEGER"}]})
        )
        (query_path / "metadata.yaml").write_text(
            yaml.dump({"scheduling": {"destination_table": None}})
        )

        with pytest.raises(
            deploy.SkippedDeployException, match="null destination_table configured"
        ):
            deploy.deploy_table(artifact_file=query_path / "query.sql")

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.dryrun.DryRun")
    def test_deploy_table_already_exists(self, mock_dryrun, mock_client, query_path):
        mock_dryrun().get_schema.return_value = {
            "fields": [{"name": "f0_", "type": "INTEGER"}]
        }

        client = mock_client.return_value
        deploy.deploy_table(
            artifact_file=query_path / "query.sql",
        )

        client.update_table.assert_called_once()
        assert client.create_table.call_count == 0

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.dryrun.DryRun")
    def test_deploy_table_already_exists_skip_existing(
        self, mock_dryrun, mock_client, query_path
    ):
        mock_dryrun().get_schema.return_value = {
            "fields": [{"name": "f0_", "type": "INTEGER"}]
        }

        with pytest.raises(deploy.SkippedDeployException, match="already exists"):
            deploy.deploy_table(
                artifact_file=query_path / "query.sql", skip_existing=True
            )

    @patch("bigquery_etl.dryrun.DryRun")
    def test_deploy_fails_when_schemas_dont_match(self, mock_dryrun, query_path):
        mock_dryrun().get_schema.return_value = {
            "fields": [{"name": "doesnt_exist", "type": "STRING"}]
        }
        with pytest.raises(deploy.FailedDeployException, match="does not match schema"):
            deploy.deploy_table(
                artifact_file=query_path / "query.sql",
            )

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.dryrun.DryRun")
    def test_force_deploy_with_not_matching_schemas(
        self, mock_dryrun, mock_client, query_path
    ):
        mock_dryrun().get_schema.return_value = {
            "fields": [{"name": "doesnt_exist", "type": "STRING"}]
        }
        client = mock_client.return_value
        client.get_table.side_effect = NotFound("table not found")

        deploy.deploy_table(artifact_file=query_path / "query.sql", force=True)

        client.create_table.assert_called_once()

    @patch("google.cloud.bigquery.Client")
    def test_deploy_metadata(self, mock_client, tmp_path):
        query_path = (
            tmp_path / "sql" / "moz-fx-data-shared-prod" / "test" / "test_query_v1"
        )
        query_path.mkdir(parents=True)
        (query_path / "schema.yaml").write_text(
            yaml.dump({"fields": [{"name": "f0_", "type": "INTEGER"}]})
        )
        (query_path / "metadata.yaml").write_text(yaml.dump({}))
        client = mock_client.return_value
        client.get_table.side_effect = NotFound("table not found")

        deploy.deploy_table(
            artifact_file=query_path / "metadata.yaml",
        )

        client.create_table.assert_called_once()
        assert client.update_table.call_count == 0



from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

import attr
import pytz

from bigquery_etl.metrics import MetricHub


@attr.s(auto_attribs=True)
class Dimension:
    data_source: str
    select_expression: str


class Aggregation(Enum):
    MIN = "min"
    MAX = "max"
    COUNT = "count"
    COUNT_DISTINCT = "count distinct"
    MEAN = "mean"

    def sql(self, column: str) -> str:
        """Returns the SQL snippet."""

        if self.value == "min":
            return f"MIN({column})"
        elif self.value == "max":
            return f"MAX({column})"
        elif self.value == "count":
            return f"COUNT({column})"
        elif self.value == "count distinct":
            return f"COUNT(DISTINCT {column})"
        elif self.value == "mean":
            return f"AVG({column})"
        elif self.value == "sum":
            return f"SUM({column})"
        elif "{column}" in self.value:
            return self.value
        else:
            raise ValueError(f"No SQL implemented for {self.value}")


@attr.s(auto_attribs=True)
class Funnel:
    steps: List[str]
    dimensions: Optional[List[str]] = attr.ib(None)


@attr.s(auto_attribs=True)
class Step:
    data_source: str
    aggregation: Aggregation
    select_expression: str
    friendly_name: Optional[str] = attr.ib(None)
    description: Optional[str] = attr.ib(None)
    where_expression: Optional[str] = attr.ib(None)
    join_previous_step_on: Optional[str] = attr.ib(None)


@attr.s(auto_attribs=True)
class DataSource:
    from_expression: str
    submission_date_column: str = attr.ib("submission_date")
    client_id_column: str = attr.ib("client_id")


def parse_date(yyyy_mm_dd: Optional[str]) -> Optional[datetime]:
    if not yyyy_mm_dd:
        return None
    return datetime.strptime(yyyy_mm_dd, "%Y-%m-%d").replace(tzinfo=pytz.utc)


def _validate_yyyy_mm_dd(instance: Any, attribute: Any, value: Any) -> None:
    parse_date(value)


@attr.s(auto_attribs=True)
class FunnelConfig:
    funnels: Dict[str, Funnel]
    steps: Dict[str, Step]
    dimensions: Dict[str, Dimension] = attr.ib({})
    data_sources: Dict[str, DataSource] = attr.ib({})
    destination_dataset: str = attr.ib("telemetry_derived")
    version: str = attr.ib("1")
    platform: Optional[str] = attr.ib(None)
    owners: Optional[List[str]] = attr.ib(None)
    start_date: Optional[str] = attr.ib(None, validator=_validate_yyyy_mm_dd)

    def __attrs_post_init__(self):
        # check if metric-hub data source was referenced
        metric_hub = MetricHub()

        for step_name, step in self.steps.items():
            if step.data_source not in self.data_sources:
                if not self.platform:
                    raise ValueError(
                        f"Undefined data source {step.data_source} for step {step_name}. "
                        + "If you are referencing a metric-hub data source, please specify the platform."
                    )
                else:
                    data_source_sql = metric_hub.data_source(
                        data_source=step.data_source, platform=self.platform
                    )
                    self.data_sources[step.data_source] = DataSource(
                        from_expression=data_source_sql
                    )

        for _, dimension in self.dimensions.items():
            if dimension.data_source not in self.data_sources:
                if not self.platform:
                    raise ValueError(
                        f"Undefined data source {step.data_source} for step {step_name}. "
                        + "If you are referencing a metric-hub data source, please specify the platform."
                    )
                else:
                    data_source_sql = metric_hub.data_source(
                        data_source=dimension.data_source, platform=self.platform
                    )
                    self.data_sources[step.data_source] = DataSource(
                        from_expression=data_source_sql
                    )

        # todo: allow referencing dimensions from metric-hub



"""Generic utility functions."""

import glob
import logging
import os
import random
import re
import string
import tempfile
import warnings
from pathlib import Path
from typing import List, Optional, Set, Tuple
from uuid import uuid4

import click
import sqlglot
from google.cloud import bigquery
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.config import ConfigLoader
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.metrics import MetricHub

# Search for all camelCase situations in reverse with arbitrary lookaheads.
REV_WORD_BOUND_PAT = re.compile(
    r"""
    \b                                  # standard word boundary
    |(?<=[a-z][A-Z])(?=\d*[A-Z])        # A7Aa -> A7|Aa boundary
    |(?<=[a-z][A-Z])(?=\d*[a-z])        # a7Aa -> a7|Aa boundary
    |(?<=[A-Z])(?=\d*[a-z])             # a7A -> a7|A boundary
    """,
    re.VERBOSE,
)
FILE_PATH = Path(os.path.dirname(__file__))
DEFAULT_QUERY_TEMPLATE_VARS = {
    "is_init": lambda: False,
    "metrics": MetricHub(),
}
ROOT = Path(__file__).parent.parent.parent
CHECKS_MACROS_DIR = ROOT / "tests" / "checks"


def snake_case(line: str) -> str:
    """Convert a string into a snake_cased string."""
    # replace non-alphanumeric characters with spaces in the reversed line
    subbed = re.sub(r"[^\w]|_", " ", line[::-1])
    # apply the regex on the reversed string
    words = REV_WORD_BOUND_PAT.split(subbed)
    # filter spaces between words and snake_case and reverse again
    return "_".join([w.lower() for w in words if w.strip()])[::-1]


def project_dirs(project_id=None, sql_dir=None) -> List[str]:
    """Return all project directories."""
    if sql_dir is None:
        sql_dir = ConfigLoader.get("default", "sql_dir", fallback="sql")
    if project_id is None:
        return [
            os.path.join(sql_dir, project_dir)
            for project_dir in os.listdir(sql_dir)
            # sql/ can contain files like bigconfig.yml
            if os.path.isdir(os.path.join(sql_dir, project_dir))
        ]
    else:
        return [os.path.join(sql_dir, project_id)]


def random_str(length: int = 12) -> str:
    """Return a random string of the specified length."""
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))


def render(
    sql_filename,
    template_folder=".",
    format=True,
    imports=[],
    **kwargs,
) -> str:
    """Render a given template query using Jinja."""
    path = Path(template_folder) / sql_filename
    skip = {
        file
        for skip in ConfigLoader.get("render", "skip", fallback=[])
        for file in glob.glob(
            skip,
            recursive=True,
        )
    }
    test_project = ConfigLoader.get("default", "test_project")
    sql_dir = ConfigLoader.get("default", "sql_dir", fallback="sql")

    if test_project in str(path):
        # check if staged file needs to be skipped
        skip.update(
            [
                p
                for f in [Path(s) for s in skip]
                for p in glob.glob(
                    f"{sql_dir}/{test_project}/{f.parent.parent.name}*/{f.parent.name}/{f.name}",
                    recursive=True,
                )
            ]
        )

    if any(s in str(path) for s in skip):
        rendered = path.read_text()
    else:
        if sql_filename == "checks.sql":
            # make macros available by creating a temporary file and pasting macro definition
            # alongside checks.
            # it is not possible to use `include` or `import` since the macros live in a different
            # directory than the checks Jinja template.
            with tempfile.NamedTemporaryFile(mode="w+") as tmp_checks_file:
                macro_imports = "\n".join(
                    [
                        macro_file.read_text()
                        for macro_file in CHECKS_MACROS_DIR.glob("*")
                    ]
                )
                checks_template = Path(str(tmp_checks_file.name))
                checks_template.write_text(
                    macro_imports
                    + "\n"
                    + (Path(template_folder) / sql_filename).read_text()
                )

                file_loader = FileSystemLoader(f"{str(checks_template.parent)}")
                env = Environment(loader=file_loader)
                main_sql = env.get_template(checks_template.name)
        else:
            file_loader = FileSystemLoader(f"{template_folder}")
            env = Environment(loader=file_loader)
            main_sql = env.get_template(sql_filename)

        template_vars = DEFAULT_QUERY_TEMPLATE_VARS | kwargs
        rendered = main_sql.render(**template_vars)

    if format:
        rendered = reformat(rendered)
    return rendered


def get_table_dir(output_dir, full_table_id):
    """Return the output directory for a given table id."""
    return Path(os.path.join(output_dir, *list(full_table_id.split(".")[-2:])))


def write_sql(output_dir, full_table_id, basename, sql, skip_existing=False):
    """Write out a query to a location based on the table ID.

    :param output_dir:    Base target directory (probably sql/moz-fx-data-shared-prod/)
    :param full_table_id: Table ID in project.dataset.table form
    :param basename:      The name to give the written file (like query.sql)
    :param sql:           The query content to write out
    :param skip_existing: Whether to skip an existing file rather than overwriting it
    """
    d = get_table_dir(output_dir, full_table_id)
    d.mkdir(parents=True, exist_ok=True)
    target = d / basename
    if skip_existing and target.exists():
        logging.info(f"Not writing {target} because it already exists")
        return
    logging.info(f"Writing {target}")
    with target.open("w") as f:
        f.write(sql)
        f.write("\n")


def qualify_table_references_in_file(path: Path) -> str:
    """Add project id and dataset id to table/view references and persistent udfs in a given query.

    e.g.:
    `table` -> `target_project.default_dataset.table`
    `dataset.table` -> `target_project.dataset.table`

    This allows a query to run in a different project than the sql dir it is located in
    while referencing the same tables.
    """
    # sqlglot cannot handle scripts with variables and control statements
    if re.search(
        r"^\s*DECLARE\b", path.read_text(), flags=re.MULTILINE
    ) or path.name in ("script.sql", "udf.sql"):
        raise NotImplementedError(
            "Cannot qualify table_references of query scripts or UDFs"
        )

    # determine the default target project and dataset from the path
    target_project = Path(path).parent.parent.parent.name
    default_dataset = Path(path).parent.parent.name

    # sqlglot doesn't support Jinja, so we need to render the queries and
    # init queries to raw SQL
    sql_query = render(
        path.name,
        template_folder=path.parent,
        format=False,
        **DEFAULT_QUERY_TEMPLATE_VARS,
    )
    init_query = render(
        path.name,
        template_folder=path.parent,
        format=False,
        is_init=lambda: True,
        metrics=MetricHub(),
    )
    # use sqlglot to get the SQL AST
    init_query_statements = sqlglot.parse(
        init_query,
        read="bigquery",
    )
    sql_query_statements = sqlglot.parse(sql_query, read="bigquery")

    # tuples of (table identifier, replacement string)
    table_replacements: Set[Tuple[str, str]] = set()

    # find all non-fully qualified table/view references including backticks
    for query in [init_query_statements, sql_query_statements]:
        for statement in query:
            if statement is None:
                continue

            cte_names = {
                cte.alias_or_name.lower() for cte in statement.find_all(sqlglot.exp.CTE)
            }

            table_aliases = {
                cte.alias_or_name.lower()
                for cte in statement.find_all(sqlglot.exp.TableAlias)
            }

            for table_expr in statement.find_all(sqlglot.exp.Table):
                # existing table ref including backticks without alias
                table_expr.set("alias", "")
                reference_string = table_expr.sql(dialect="bigquery")

                matched_cte = [
                    re.match(
                        rf"^{cte}(?![a-zA-Z0-9_])",
                        reference_string.replace("`", "").lower(),
                    )
                    for cte in cte_names.union(table_aliases)
                ]
                if any(matched_cte):
                    continue

                # project id is parsed as the catalog attribute
                # but information_schema region may also be parsed as catalog
                if table_expr.catalog.startswith("region-"):
                    project_name = f"{target_project}.{table_expr.catalog}"
                elif table_expr.catalog == "":  # no project id
                    project_name = target_project
                else:  # project id exists
                    continue

                # fully qualified table ref
                replacement_string = f"`{project_name}.{table_expr.db or default_dataset}.{table_expr.name}`"

                table_replacements.add((reference_string, replacement_string))

    updated_query = path.read_text()

    for identifier, replacement in table_replacements:
        if identifier.count(".") == 0:
            # if no dataset and project, only replace if it follows a FROM, JOIN, or implicit cross join
            regex = (
                r"(?P<from>(FROM|JOIN)\s+)"
                r"(?P<cross_join>[a-zA-Z0-9_`.\-]+\s*,\s*)?"
                rf"{identifier}(?![a-zA-Z0-9_`.])"
            )
            replacement = r"\g<from>\g<cross_join>" + replacement
        else:
            identifier = identifier.replace(".", r"\.")
            # ensure match is against the full identifier and no project id already
            regex = rf"(?<![a-zA-Z0-9_`.]){identifier}(?![a-zA-Z0-9_`.])"

        updated_query = re.sub(
            re.compile(regex),
            replacement,
            updated_query,
        )

    # replace udfs from udf/udf_js that do not have a project qualifier
    regex = r"(?<![a-zA-Z0-9_.])`?(?P<dataset>udf(_js)?)`?\.`?(?P<name>[a-zA-Z0-9_]+)`?"
    updated_query = re.sub(
        re.compile(regex),
        rf"`{target_project}.\g<dataset>.\g<name>`",
        updated_query,
    )

    return updated_query


def extract_last_group_by_from_query(
    sql_path: Optional[Path] = None, sql_text: Optional[str] = None
):
    """Return the list of columns in the latest group by of a query."""
    if not sql_path and not sql_text:
        raise click.ClickException(
            "Extracting GROUP BY from query failed due to sql file"
            " or sql text not available."
        )

    if sql_path:
        try:
            query_text = sql_path.read_text()
        except (FileNotFoundError, OSError):
            raise click.ClickException(f'Failed to read query from: "{sql_path}."')
    else:
        query_text = str(sql_text)
    group_by_list = []

    # Remove single and multi-line comments (/* */), trailing semicolon if present and normalize whitespace.
    query_text = re.sub(r"/\*.*?\*/", "", query_text, flags=re.DOTALL)
    query_text = re.sub(r"--[^\n]*\n", "\n", query_text)
    query_text = re.sub(r"\s+", " ", query_text).strip()
    if query_text.endswith(";"):
        query_text = query_text[:-1].strip()

    last_group_by_original = re.search(
        r"(?i){}(?!.*{})".format(re.escape("GROUP BY"), re.escape("GROUP BY")),
        query_text,
        re.DOTALL,
    )

    if last_group_by_original:
        group_by = query_text[last_group_by_original.end() :].lstrip()
        # Remove parenthesis, closing parenthesis, LIMIT, ORDER BY and text after those. Remove also opening parenthesis.
        group_by = (
            re.sub(r"(?i)[\n\)].*|LIMIT.*|ORDER BY.*", "", group_by)
            .replace("(", "")
            .strip()
        )
        if group_by:
            group_by_list = group_by.replace(" ", "").replace("\n", "").split(",")
    return group_by_list


class TempDatasetReference(bigquery.DatasetReference):
    """Extend DatasetReference to simplify generating temporary tables."""

    def __init__(self, *args, **kwargs):
        """Issue warning if dataset does not start with '_'."""
        super().__init__(*args, **kwargs)
        if not self.dataset_id.startswith("_"):
            warnings.warn(
                f"temp dataset {self.dataset_id!r} doesn't start with _"
                ", web console will not consider resulting tables temporary"
            )

    def temp_table(self) -> bigquery.TableReference:
        """Generate a temporary table and return the specified date partition.

        Generates a table name that looks similar to, but won't collide with, a
        server assigned table and that the web console will consider temporary.

        In order for query results to use time partitioning, clustering, or an
        expiration other than 24 hours, destination table must be explicitly set.
        Destination must be generated locally and never collide with server
        assigned table names, because server-assigned tables cannot be modified.
        Server assigned tables for a dry_run query cannot be reused as that
        constitutes a modification. Table expiration can't be set in the query job
        config, but it can be set via a CREATE TABLE statement.

        Server assigned tables have names that start with "anon" and follow with
        either 40 hex characters or a uuid replacing "-" with "_", and cannot be
        modified (i.e. reused).

        The web console considers a table temporary if the dataset name starts with
        "_" and table_id starts with "anon" and is followed by at least one
        character.
        """
        return self.table(f"anon{uuid4().hex}")



"""Extract Thompson sampling prior query results and write the combined JSON to a single file."""

from bigquery_etl.newtab_merino import export_newtab_merino_table_to_gcs

if __name__ == "__main__":
    export_newtab_merino_table_to_gcs()
